/*
 *  this is a collection of diagonalization code used by spinpack
 */

#include "mem.h"
#include "spinstri.h"
#include "spins.h"
#include "block.h"
#include "vector.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif

double ta[MAX_ITR];
double tb[MAX_ITR];		/* a,b Tridiagonalmatrix		*/
double tc[MAX_ITR][NEV];		/* eigenvectors of tridiagonalmatrix	*/
    
/* --- --- get smallest eigenvalues by Lanczos (Gagliano) ----------   
  - projections vector |Pi> (start vector |P0>)
  - 2nd Ebenenvektor |Pis>=((H-<H>_i)|Pi>)/(<H^2>_i-<H>_i^2)^(1/2)
    perpenticular to |Pi> (<Pi|Pis>=0)
  - compute 2x2-matrix for |Pi> and |Pis> and their EW
  - |Pi+1>=a|Pi>+b|Pis> is normalized EV to lower EW

  v0=|Pi> v1=H|Pi> v2=HH|Pi>
  h=<H>_i hh=<HH>_i hhh=<HHH>_i
  b=q(<H^2>_i-<H>_i^2)=q(hh-h^2)          mit  q(x)=x^(1/2)
  H-Matrix=[[<Pi|H|Pi>,<Pi|H|Pis>],[<Pis|H|Pi>,<Pis|H|Pis>]]
          =[[h,b],[b,(hhh-2*hh*h+h^3)/b^2]]
          =[[h,b],[b,(hhh-h*hh)/b^2-h]]
          =[[h,b],[b,2*b*f+h]]
  EW=(a11+a22)/2-q((a11+a22)^2/4-a11*a22+a12^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*((hhh-h*hh)/b^2-h)+b^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*(hhh-h*hh)/b^2+hh)
    =h+b*f-q((b*f+h)^2-2*h*b*f-h^2+b^2) ,    a22=2*b*f+h
    =h+b*f-q(b^2*f^2+b^2)
    =h+b*(f-q(1+f*f))                   ,    alf=f-q(1+f*f)
    =h+b*alf
  EV=[(EW-a22)/a12,1]/q((EW-a22)^2/a12^2+1) * [Pi,Pis]
    =((EW-a22)/b*v0+(1/b*v1-h/b*v0))/q((EW-a22)^2/b^2+1)
    =((EW-a22-h)*v0+v1)/q((EW-a22)^2+b^2)
  f=(a22-a11)/(2*a12),     -p/2=(a11+a22)/2=a22-a12*f=a12*f+a11=b*f+h
   --- last change 7.12.98 JS ueberarbeitet ---
 * --- --- -------------------------------------------------- ---- --- */
/* --- Inp: v0  Out: v0,itr,ew  Wrk: v1 */
int lanz1(tvec *v0,tvec *v1,tvec *ev){
    tvec *v2;     cplx a, alf, ak, b, f, r, h, hh, hhh;
    int itr, j;   ULong i;   double eps=1e-6;	
    /* eps=precission^2 = <(H-<H>)^2> = <HH>-<H><H> >= 0, also complex ??? */
    v2=ev;		/* new 12.09.1998 use memory of ev[n] */
    r=ak=alf=f=0;
    for(itr=0;;itr++){
      clrvec(v1);hamilton(v0,v1);		/* v1=H*v0+v1 */
      clrvec(v2);hamilton(v1,v2);
      for(hhh=hh=h=0,i=0l;i<b_ofs[pt_n];i++){
        r=Conj(M_f2d(v0[i]));
        h += r*M_f2d(v1[i]);
        hh+= r*M_f2d(v2[i]);
        hhh+=Conj(M_f2d(v1[i]))*M_f2d(v2[i]); }
      b=hh-h*h;		/* negativ if rounding-errors! cmplx>0 ??? */
      b=fabs(b);	/* if cmplx (gegen 0) */
      if(D(b)>0){ /* new projection vector perpenticular to the old one */
        b=sqrt(b); 				/* 2x2 diagonalization */
        f=(hhh-3*h*hh+2*h*h*h)/(2*b*b*b);	/* gegen inf */
        alf=f-sqrt(1.+f*f);			/* gegen -0 bis f*f OVL */ 
        a=1./sqrt(1.+alf*alf);			/* gegen  1 */
        ak=a*alf/b;				/* gegen -0 */
        r=a-ak*h;				/* gegen  1 */
        for(i=0l;i<b_ofs[pt_n];i++)
        v0[i]=M_d2f(r*M_f2d(v0[i])+ak*M_f2d(v1[i]));
        /* h=h+b*alf; */ }
      printf("\n i=%3d h=%16.12f alf=%.4e b=%.1e v=[% .3f,% .3f] t+=%s",
             itr,D(h),D(alf),D(b),D(r),D(h*ak),sclock(~0l));
      if(/* itr>1 && */ fabs(b)<=eps) break; }
    for(i=0l;i<b_ofs[pt_n];i++) ev[i]=v0[i];
    for(j=0;j<NEW;j++)ew[j]=D(h);
    return(itr); }
      
/* ----------- bisection method -------------------------------------
 * calculate the eigenvalues of a tridiagonal matrix T=(tb,ta,tb)
 *  Inp: ta,tb=Tridiagonalmatrix(T),  nn=size of the matrix T
 *       ne=number of eigenvalues ew to calculate, ep=precision
 *  Out: ew  eigenvalues of T via "Nullstellen" of Det(A-x*I)
 *  What happens for degeneracy?
 *    => EW only one-fold, but should be n-fold (as expected by vec12)
 *    ToDo: after EW[ne] is converged, call another exact/slow tridiag
 *          p.e. dsterf() of lapack/icepack ? t(300)=126s acceptable
 *          sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2 
 *      OR try (modified) vec12 morefold on same eigenvalue until <tc>=0
 *      check on til8_18b.dat
 * add ta[1..nn-1],tb[0..nn-2] as parameter 2012-05
 * --- */
void bisec(int nn,int ne,double ep, double *ta, double *tb){
  static double b2[MAX_ITR]; /* if not enough stackspace, faster */
  double ra,rb,rc,g,epa,range; int i,j,k,neg,ipass;
  /* find a interval containing all the eigenvalues rb..ra */
  range=fabs(ta[0])+fabs(tb[0]);
  for(k=1;k<nn-1;k++) range=max(range,fabs(tb[k-1])+fabs(ta[k])+fabs(tb[k]));
  range=max(range,fabs(tb[nn-2])+fabs(ta[nn-1])); range=-range;
  /* range = max_i (sum_j fabs(matrix_ij))  matrix line */
  b2[0]=0; for(i=1;i<nn;i++) b2[i]=Norm(tb[i-1]);
  epa=fabs(range*ep); for(i=0;i<ne;i++) ew[i]=-range;
  rb=range;
  /* find the eigenvalues of a tridiagonal matrix * 
   * lying in a given interval rb...ra            */
  for(k=0;k<ne;k++) { ra=ew[k];       /* ra=upper bound, rb=lower bound */
    /* was j<100, but this is a bit to small ??? */
    for(j=0;j<10000;j++){		/* sturm sequencing */
      if(fabs(ra-rb)<epa)break;	/* ra=rb => break */
      rc=(ra+rb)*0.5; ipass=neg=0; g=1.0;
      for(i=0;i<nn;i++){   /* g=Det_{i}(rc)/Det_{i-1}(rc) */
        if(ipass==0)g=rc-ta[i]-b2[i]/g;else	/* do not divide if g=0 */
        if(ipass==1)ipass=2;
        else      { ipass=0; g=rc-ta[i]; }	/* new start, decoupled */
        if(ipass==0){ if(g<0.0)neg++;		/* sign-change of Det */
          if(fabs(g)<=fabs(b2[i]*epa*ep)) ipass=1; } /* g==0 ? */
      }   
      neg=nn-neg;   /* num of nullstellen rc...ra -> rb...rc */
      if(neg<=k)rb=rc;		/* new lower bound */
      else    { ra=rc; 		/* new upper bound */
                for(i=k;i<min(neg,ne);i++)ew[i]=rc; }
    } }
  return; }

/* --- eigenvectors (EV) of a tridiagonal matrix by inverse iteration
 *  http://en.wikipedia.org/wiki/Power_iteration Mai2010
 *     inverse iteration == power iteration for inverse matrix (A^(-1))
 *                          + choice to choose the eigenvalue to converge to
 *  see Numerical Recipes  (Feb99)
 *   A - matrix
 *   a - start-eigenvalue
 *   x - startvector (random)
 *   step1: solve equation system (A-a*1)*y==x for vector y (ex: LU decomposition)
 *   step2: x := y/|y|          (normalize)
 *   step3: a := a+(x,x)/(x,y)  (not necessary every step)
 *   if Ax=ax then break else goto step1   
 *
 * --- nn i Matrixgroesse, {ta,tb} TriDiagMatrix -> tc EVs
 * --- for degeneracy EW must be n-fold in ew[] and nev>=n!  (does not work!)
 *   ex: ew[]={ -3.5, -3.5, -3.5, -1.5} for 3-fold degeneracy
 *   input: nn - number of sites, nev - number eigenvectors,
 *          ne0 - offset for lowest eigenvectors (0..n1-NEV)
 */
double di[MAX_ITR],bl[MAX_ITR],bu[MAX_ITR],
     bv[MAX_ITR],cm[MAX_ITR];
char lex[MAX_ITR];	/* 0 (non pivoting) or 1 (pivoting) */
void vec12(int nn, int nev, int ne0){
    double s,prd; int i,j,k,km,l;
    km=0;/* km=0 added by JS */
    for(k=0;k<nev;k++){ /* if ew0==ew1==ew2 : 3 orthogonal vectors */
        for(j=0;j<nn;j++) {di[j]=ew[k+ne0]-ta[j]; bl[j]=bu[j]=-tb[j];}
        /* --- LU decomposition, prepare lex, cm, bu, bv, di */ 
        for(j=0;j<nn-1;j++) {
           if(fabs(di[j])>fabs(bl[j])){		/* --- non pivoting */
             lex[j]=0; if(fabs(di[j])<1e-30) di[j]=1e-30; /* e-40 => e-30 (float) */
             cm[j+1]=bl[j]/di[j]; di[j+1]-=cm[j+1]*bu[j]; bv[j]=0; }  
           else {				/* --- pivoting */
             lex[j]=1; cm[j+1]=di[j]/bl[j]; di[j]=bl[j]; s=bu[j];
             bu[j]=di[j+1]; bv[j]=bu[j+1]; di[j+1]=s-cm[j+1]*bu[j];
             bu[j+1]=-cm[j+1]*bv[j]; }
        }  
        if(fabs(di[nn-1])<1e-30) di[nn-1]=1e-30;
        
        /* --- initial vec tc */ 
        for(j=0;j<nn;j++) tc[j][k]=1./((j+1)*5.); /* random ??? */
        /* --- degeneracy check up -> Entartung km..k */
        /*  degeneracy is given in ew[] implicitely!  */
        if(k==0) km=k; else 
         if(fabs(ew[k+ne0]-ew[km+ne0])>1e-10) km=k;     /* old=1e-13 1998 */ 
         else for(i=km;i<k;i++){       /* orthogonalise ini_vec to tri_ev */
            for(prd=0,j=0;j<nn;j++) prd+=tc[j][i]*tc[j][k];  /* <c_i|c_k> */
            for(      j=0;j<nn;j++) tc[j][k]-=prd*tc[j][i]; }
        /* ToDo: stop and ignore EW, if <tc|tc>==0 ? */
        /* --- inverse iteration of tc (inp: lex, cm, bu, bv, di) */
        for(l=0;l<(k-km+3);l++){	/* what does l??? */
            if((l!=0)||(k!=km)){ /* --- forward substitution */
               for(j=0;j<nn-1;j++)
               if(lex[j]==0) tc[j+1][k]-=cm[j+1]*tc[j][k];
               else {s=tc[j][k];tc[j][k]=tc[j+1][k];
                     tc[j+1][k]=s-cm[j+1]*tc[j][k];}
            }  /* --- backward substitution */
            for(j=nn-1;j>=0;j--){
                s=tc[j][k];
                if(j<(nn-1))s-=bu[j]*tc[j+1][k];
                if(j<(nn-2))s-=bv[j]*tc[j+2][k];
                tc[j][k]=s/di[j];
            }   /* --- normalization */
            for(prd=0,j=0;j<nn;j++) prd+=Norm(tc[j][k]);
            if(prd>1e-30)prd=1./sqrt(prd);else prd=0;
            for(j=0;j<nn;j++) tc[j][k]*=prd;
        }
    }   return;
}

/*
 * the degeneracy from ta,tb is smaller than the degeneracy of H
 *         (and the same as got by fulltridiag)
 * probably reason:
 *   we start with a random vector which defines only a subspace if H has
 *    degenerated eigenvalues
 *   the other degenerated states are natural (mostly or always?) orthogonal
 *    to this space by definition
 *   that should be a problem for all iterating algorithms
 */

/* ew  - eigenvalues of iteration i
 * et  - eigenvalues of iteration i-1
 * eps - accuracy (1e-5)
 * - Mai99 - iteration stops to late
 * - Mar04 - rarely iteration stops to early
 * - Feb09 - problems with FM-square-J1J2-float e0-not conv, but e2 conv.
 *  recommend pew < NEW-1
 * out: ewc - list of converged ergs
 */
double ewc[NEW];
int convergence(double *ew, double *et, double eps, int itr){	/* see inc1() */
  int i1, i2, i3, i4=0, num_conv=0, in=min(max(1,pew),NEW);
  for (i3=i1=0;i1<((!num_conv)?1:min(itr-1,NEW));i1++){ /* check all values */
                 /* ^^^^^^^^^^^^ ew[0] must be converged first! (FM-j1j2) */
    for(i2=i3;i2<=i1;i2++)
      if(fabs(et[i2]-ew[i1])<=fabs(eps*ew[i1])){ /* found same value */
        ewc[num_conv]=ew[i1]; /* store converged values only for awk/grep */
        num_conv++;   /* count convergent values */
        i3=i2; i4=i1; /* avoid looking twice */
        break;
      } 
    if (i1>i4+4) break; /* to much nonconverged values */
  }
  if (n1<=(unsigned)itr) { /* for small n1 we have no soft convergence */
    num_conv = min((int)n1,NEW); i4=num_conv-1;
    for(i1=0;i1<num_conv;i1++) ewc[i1]=ew[i1]; /* converged values */
  }
  for(i1=num_conv;i1<NEW;i1++) ewc[i1]=ew[i4]; /* fill with last converged value */
  for(i1=0;i1<NEW;i1++) et[i1]=ew[i1];
  /* ToDo: save converged values in ew_conv? for better awk/grep of E0 E1 ... */
  if (!mpi_i) printf(" cnv= %d %d",num_conv,i4);
  if (i4==NEW-1 && num_conv<in && !mpi_i)
    printf ("\nWarning: (NEW) to small to get (pew) values.");
  return (num_conv>=in || i4==NEW-1);
}

void fulltridiag(unsigned int n1,double ta[], double tb[]);

#define DEBUG 1
/* on job-systems, we have a walltime limit, we can use exit_it to
 * limit the computing time by checkpoint+exit and resume later
 */
int chkpt_time= 4*24*3600; // seconds between checkpoints dflt= 4d
/* --- --- get smallest eigenvalues (ew[]) using Lanczos --- */
/* --- Inp: iv  Out: a,b,itr,ew  Wrk: v0,v1 ret: Iterationen */
int inc1(int nev, int ne0, tvec *v0, tvec *v1){
  int      chkpt_now= 0; // set checkpoint now?
  int      chkpt_last_i= 0;    // iteration of last timestamp
  double   chkpt_last_t= 0.0;  // time in seconds of last timestamp
  int i=0, i_start=0 /* 0(ok) or 1(bad) ? 2013-04 answered by tests */, k; 
  double tm1, tm2=0, tm2b=0, tm3;  /* measure time and speed */
  double a1, b1, eps, et[NEW]; cplx t1; tvec *vtmp;
  if (vvv && !mpi_i)
    printf("\n --- starting Lanczos diagonalization ---"
           "\n inc1-lanczos ne0=%d (out=tmp/tri.txt)", ne0);
  clrvec(v0);
  if (vvv && !mpi_i) printf(" t+= %s",sclock(~0l));
  if (!mpi_i)
    printf("\n start iteration: E0..E%d i i100.t[min] num_conv last_conv",
          NEW-1);
  for (k=0;k<NEW;k++) ew[k]=et[k]=13.5;
#if VecType & 1
  eps=1e-9;/* high precision 1e-11 */
#else
  eps=1e-9;/* low precision 1e-9 */
#endif
  MPI_Barrier(MPI_COMM_WORLD); /* sync outputstreams DBG */
  tm1= fstime();
  i=1; /* start iteration with 1 */
  // checkpoint=6 = inc1 = ew-iteration
  // checkpoint=7 = ew-iteration finished
  // checkpoint=8 = inc12 = ev-iteration
  // checkpoint=9 = ev-iteration finished
  if (checkpoint>=6) { // old: 4 JS 2012-05-10 (test for mpi_n=72 maxfile=0)
    /* reload old values */
    i= loadtri(ta,tb,MAX_ITR); /* all threads, ta[1..nn-1],tb[0..nn-2] */
    if (n1<1000 && !mpi_i) { int j; printf("\n"); // debugging only, small n1
      for (j=0;j<=i && j<MAX_ITR;j++)
        printf(" i= %3d % f % f\n",j,D(ta[j]),D(tb[j]));
    }
    if (!mpi_i) printf("\n inc1.chkpt6.resume loadtri= %d", i);
    bisec(i,NEW,1e-13,ta,tb);
    if(!mpi_i){nl(stdout);outew(stdout);printf(" i= %2d chkpt6.loadtri\n",i);}
    // ToDo:  only load if chkpt6 is loadable
    // if (checkpoint==6) i++; // correct iteration index to next = start loop
    if (checkpoint<7) { // 2012-05-10 JS, may have chkpt7 without any chkpt6
      int i6=-1; // i from checkpoint6 and tri-file may differ by one
      // -1 is default = error read chkpt6
      // first look at chkpt6 or older chkpt6.old
      if (!mpi_i){ // read last checkpointed iteration state, default i=1
        i6=i;
        FILE *f1; f1=fopen(FCHKPT"6","rt");
        if (!f1/* && errno==ENOENT*/) { // try rename .old and try again
          printf(" WARNING: no chkpt6, restoring chkpt6.old ...\n");
          // JS201209 better use odd/even to avoid critical phase
          // rename("tmp_shared/v0.old","tmp_shared/v0.dat"); // unistd.h
          // rename("tmp_shared/v1.old","tmp_shared/v1.dat"); // unistd.h
          rename(FCHKPT"6.old",FCHKPT"6");
          sleep(5); // relax before probably accessing from other nodes?
          f1=fopen(FCHKPT"6","rt");
        }
        if (!f1) F0("no valid chkpt6[.old], please try to reduce chkpt + restart");
        // possible new i <= old i !!!
        if (f1) if (fscanf(f1,"%d",&i6)) {
            /* i++; bad! */ printf(" inc1.chkpt6.load i= %d", i6);
          }
        if (f1) fclose(f1);
      }
      // ToDo: test chpt5 which means that step4=inc1 is finished
      MPI_Bcast(&i,1,MPI_INT,0,MPI_COMM_WORLD); // destsrc,num,type,from,...
      if (i&1) {  // even/odd(i)
        if (loadvec(v0,"tmp_shared/v0o.dat",-1,-1)) i6=-1;
        if (loadvec(v1,"tmp_shared/v1o.dat",-1,-1)) i6=-1;
      } else {
        if (loadvec(v0,"tmp_shared/v0e.dat",-1,-1)) i6=-1;
        if (loadvec(v1,"tmp_shared/v1e.dat",-1,-1)) i6=-1;
      }
      if (i!=i6 && !mpi_i)
        printf("Warning: chpt6.i=%d != chkpt6.tri.i=%d",i6,i);
      i=i6;
      i++;  /* 20120927 shifted to this position, because loop start is 1 */
    }
    MPI_Barrier(MPI_COMM_WORLD);
  }
  /* 2013-04-25 start from 0 (i-1) for correct extrapolations */
  i_start=i-1; /* for better runtime estimation after checkpoint.resume */
  chkpt_last_t= fstime();  // needed to estimate the next checkpoint
  chkpt_last_i= i-1;
  if (checkpoint < 7) { /* old 2012-05-10: checkpoint < 4 */
    // Warning: start index = 1 (this is not C default (change?)
    for(/* i=1 */;i<=MAX_ITR;i++){  /* ToDo: start with 0 (fortran=1) */
#ifdef DEBUG
      if (vvv&256 && i<4)
      {  int mi; unsigned j;
         MPI_Barrier(MPI_COMM_WORLD);
         for (mi=0;mi<mpi_n;mi++) {  /* sort by mpi */
           if (mi==mpi_i)
           for (j=0;j<b_ofs[pt_n] && j<32;j++)
              printf("\n itr= %2d mpi= %2d i= %3lu v0_i=%12.5g v1_i=%12.5g",
                 i, mpi_i, node_ofs+j, M_f2d(Re(v0[j])), M_f2d(Re(v1[j])));
           MPI_Barrier(MPI_COMM_WORLD);
         }
         if (!mpi_i)
         printf("\n v0=H*v1+v0 v0+=-<v1Hv1>*v1 b1=<v0> v0*=1/b1 v1=-b1*v1 v1<=>v0");
         MPI_Barrier(MPI_COMM_WORLD);
      }
#endif
      /* ToDo: do the normalization within H*v operation! v1[]*=r1;v0[]*=r0 */
      /* t1=<v1*H*v1>,v0=H*v1+v0, <v1|v0>==0 !def! */
      t1=hamilton(v1,v0);      /* #FLOP see hamilton_geth_block */
      ta[i-1]=a1=Re(t1);b1=0;	/* a1=<v1*H*v1> */
      b1=add_av(v0,-a1,v1);    /* v0+=-a1*v1, b1=<v0|v0> FLOPS=4*n1 */
#ifdef DEBUG
      if (vvv&256 && i<4)
      {  int mi; unsigned j;
         MPI_Barrier(MPI_COMM_WORLD);
         for (mi=0;mi<mpi_n;mi++) {  /* sort by mpi */
           if (mi==mpi_i)
           for (j=0;j<b_ofs[pt_n] && j<32;j++)
              printf("\n itr= %2d mpi= %2d i= %3lu v0_i=%12.5g v1_i=%12.5g",
                 i, mpi_i, node_ofs+j, M_f2d(Re(v0[j])), M_f2d(Re(v1[j])));
           MPI_Barrier(MPI_COMM_WORLD);
         }
         if (!mpi_i)
         printf("\n v0=H*v1+v0 v0+=-<v1Hv1>*v1 b1=<v0> v0*=1/b1 v1=-b1*v1 v1<=>v0");
         MPI_Barrier(MPI_COMM_WORLD);
      }
#endif
      tb[i-1]=b1=sqrt(b1);     /* if b1 is 0 or near 0, we should stop */
      /* or store normalization factor seperately ? */
      if (b1>0) { /* include this in hamilton like v1=H*(v0*=a)+(v1*=b) */
        mltply_av(1.0/b1,v0); /* v0=1/b1*v0 FLOPS=1*n1 */
        mltply_av(   -b1,v1); /* v1= -b1*v1 */
      }
      vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
      /* write ta[0..i-1] tb[0..i-1] (part of diagmatrix) ?? */
      if (!mpi_i) outtri(i,a1,b1); 
      IFcplx(if(fabs(Im(t1))>1e-6)E1("Im<H>=% .0e",D(Im(t1)));) /* flt=1e-6 */

      // ToDo: checkpoint all 2h (recompute + mpibroadcast if now) i_chkpt
      //       + t_chkpt (may be killed to early by job-manager)
      //  chkpt_period // seconds
      chkpt_now= 0; // default is no checkpointing now
      if ((fstime()-chkpt_last_t) * (i+1 - chkpt_last_i) 
                    >= chkpt_time * (i+0 - chkpt_last_i)) {
        chkpt_now= 1;
      }
      /* emergency stop by signal: 2*USR2+USR1 */
      /* slurm: scancel -s USR2 (+ USR2 + USR1) $JobId # send to all threads */
      if (sig1==2) chkpt_now= 1; // we bcast it, so we dont need SIG all
      MPI_Bcast(&chkpt_now,1,MPI_INT,0,MPI_COMM_WORLD); // dat,sz,typ,snode,..

      /* "(i>19)"  removed, we can neglect the CPU-time for bisec() */
      if ( (i>0 && (i%5)==0)          /* convergence test all 5 iterations */
        || (ULong)i>=n1 || b1<1e-15 /* also stop on small systems */
        || n1<10 /* control convergence on small systems */
        || chkpt_now) {  /* output e0 state before checkpinting */
         /* compute the eigenvalues of the tridiagonal matrix T=(tb,ta,tb) */
         bisec(i,NEW,1e-13,ta,tb); 
         if (!mpi_i) {
           tm2=fstime()-tm1; /* if (tm2==0.0) tm2=0.5; */ /* seconds 1s/2 ??? */
           /* 2013-04 for autodiff outputs does not output time on every line */
           /*   should be relatively constant (for const. mem/disk accesstime) */
           /*   output 6.25% changes only (happens for moving threads = bad) */
           if (((fabs(tm2/i-tm2b)>tm2b*0.0625 && tm2>3) || i==5) && !mpi_i) {
             printf("\nPRF: iteration time: i=%3u t[m]=%7.3f i100.t[m]= %.3f",
                     i, tm2/60, (tm2*100./(i-i_start))/60.); fflush(stdout);
             tm2b=tm2/i; /* performance output */
           }
           nl(stdout); outew(stdout);
           /* move to tri.txt for better debugging using diff? */
           /* printf(" i=%3u i100=%4.0f", i, (tm2*100./(i-i_start))/60.); */
           printf(" i=%3u", i);
         }
         if (convergence(ew,et,eps,i)) break; /* print cnv=, set ewc */
         if (b1<1e-15) {if(!mpi_i)printf("\n inc1_b<1e-15"); break;}
         if ((ULong)i>=n1) break;
         if (i>=MAX_ITR) {if(!mpi_i)printf("\n no convergence"); break;}
      }
      if (chkpt_now) { /* write a checkpoint */
        tm3=fstime();
        if (!mpi_i) /* 2013-04 change itime to fstime */
          printf("\n trigger checkpointing"
                 " time[m](i=%d-%d)= %.2f est[m](%d)= %.2f", 
           i, chkpt_last_i, (tm3-chkpt_last_t)/60., i+1,
           /*chkpt_last_t+*/(tm3-chkpt_last_t)/60.*(i+1 - chkpt_last_i)
                                                  /(i   - chkpt_last_i));
        if (!mpi_i)
          printf("\n inc1.chkpt6 i=%d t[m]= %.2f", i, (fstime()-tm1)/60.);
#if 0  // remove this bad code after 2012-09! its slow and critical to break
        if (!mpi_i) { // 2012-05-22 save old chkpt (fast) 2012-09 No! to slow
          rename("tmp_shared/v0.dat","tmp_shared/v0.old"); // unistd.h
          // if (i==9 || i==9 || i==23) exit(1);
          rename("tmp_shared/v1.dat","tmp_shared/v1.old"); // unistd.h
        }
#endif
        MPI_Barrier(MPI_COMM_WORLD); // do not start before rename complete
        // slow operation, if it failes (fresh) chpt would be lost
        // by switching odd/even, we have still the former checkpoint 20120927
        if (i&1) { // even/odd(i)
          savevec(v0,"tmp_shared/v0o.dat",-1,-1);
//          if (i==9) exit(1); // test-run
          savevec(v1,"tmp_shared/v1o.dat",-1,-1);
        } else {
          savevec(v0,"tmp_shared/v0e.dat",-1,-1);
          savevec(v1,"tmp_shared/v1e.dat",-1,-1);
        }
        if (!mpi_i) { // write last checkpointed iteration state
          // in case of failed checkpointing, old files can be rescued
          rename(FCHKPT"6",FCHKPT"6.old");
          FILE *f1; f1=fopen(FCHKPT"6","wt");
          // if (f1) fprintf(f1,"6\n%lu\n%d\n",n1,i);
          if (f1) fprintf(f1,"%d\n",i);
          if (f1) fclose(f1);
          // echo 6 > tmp_shared/chkpt
          if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("6\n",f);fclose(f);}
        }
        tm2=fstime()-tm3;
        // break;
        // checkpoint 6 = chkpt6
        chkpt_last_t= fstime();
        chkpt_last_i= i;
        checkpoint=6;  // old 4
        if (n1<1000 && !mpi_i) { int j;
          printf("\n");
          for (j=0;j<=i && j<MAX_ITR;j++)
            printf(" i= %3d % f % f\n",j,D(ta[j]),D(tb[j]));
        }
        if (!mpi_i) printf(" chkpt.t=%.0fs", tm2);
        // ToDo: if (sig1) 
        if (sig1) MPI_Abort(MPI_COMM_WORLD,0);
      }
    } /* for(itr,...) */
    MPI_Barrier(MPI_COMM_WORLD); /* sync outputstreams DBG */
    if (!mpi_i) outtri(-1,0,0); /* close files */
  } /* checkpoint < 6 */
  // checkpoint 7 old 5
  if (checkpoint<7) {
    checkpoint=7; 
    if (!mpi_i) { // write last checkpointed iteration state
      // in case of failed checkpointing, old files can be rescued
      printf("\n checkpoint7.w");
      rename(FCHKPT"7",FCHKPT"7.old");
      FILE *f1; f1=fopen(FCHKPT"7","wt");
      if (f1) fprintf(f1,"%d\n",i);
      if (f1) fclose(f1);
      if(!mpi_i){FILE *f;f=fopen(FCHKPT,"wt");fputs("7\n",f);fclose(f);}
    }
  } else { // checkpoint >= 7 == load
    // ToDo: read tridiag + compute ew, used from chkpt6 loadtri ???
    if (!mpi_i) printf("\n checkpoint7.r (ToDo)");
  }
  /* ToDo: vec12 should be called by incv1, because output is needed only there
   *   for n-fold degeneracy: we need max_ea>=n, nev>=n and e0 n-fold in ew[]
   */
#if 0 /* test */
  fulltridiag(i,ta,tb); /* finds exactly the same eigenvalues! */
#endif
  if (!mpi_i) { /* output converged energies for easier awk handling v2.27 */
    printf("\n conv=");
    for (k=0;k<NEW;k++) printf(" %12.8f", ewc[k]);
    tm2=fstime()-tm1;
    printf("\nPRF: i100.t= %5.2f i_speed= %.2e hnz/s i=%3u t[m]= %.3f\n",
      (tm2*100./(i-i_start))/60., h_nz*1.*(i-i_start)/((tm2)?tm2:1), i, tm2/60.);
    if (vvv) printf(" i0= %d t+= %s", i_start, sclock(~0l));
  }
// exit(3);
  if (nev) vec12(i,nev,ne0); /* get tridiagonal eigenvectors */
  return(i);
}
    
/* --- eigenvector by the Lanczos method (after vec12())
 * --- convert eigenvectors of tridiagonal matrix back to H space 
 * same as inc1, but now we have eigenvalues (c) of triag (a,b)
 *               and do a an aditional ev[k]+=c[i][k]*v0[i]
 * --- Inp: a,b,c --- Out: v0,v1,ev=linear_access(3+itr) */
void incv1(int nev, tvec *v0, tvec *v1, tvec *ev[], int itr) {
    ULong ii=1l; int i,k; long l1; double a1,b1,rb,rc; tvec *vtmp;
    if (nev<1) return;
    if (!mpi_i) printf("\n incv1-lanczos");
    clrvec(v0);
    for (k=0;k<nev;k++) {
      copy_av(ev[k],tc[0][k],v1); /* ev[k]=tc[0][k]*v1 */
    }
    l1=itime();
    for(i=1;i<itr;i++,ii++){ /* v1.7.8 99% time spend for hamilton() */
#ifdef DEBUG
       if (vvv&256 && i<4)
       {  int mi; unsigned j;
          MPI_Barrier(MPI_COMM_WORLD);
          for (mi=0;mi<mpi_n;mi++) {  /* sort by mpi */
            if (mi==mpi_i)
            for (j=0;j<b_ofs[pt_n] && j<32;j++)
               printf("\n itr= %2d mpi= %2d i= %3lu v0_i=%12.5g v1_i=%12.5g",
                  i, mpi_i, node_ofs+j, Re(v0[j]), Re(v1[j]));
            MPI_Barrier(MPI_COMM_WORLD);
          }
          if (!mpi_i)
          printf("\n v0=H*v1+v0 v0+=-<v1Hv1>*v1 b1=<v0> v0*=1/b1 v1=-b1*v1 v1<=>v0");
          MPI_Barrier(MPI_COMM_WORLD);
       }
#endif
       hamilton(v1,v0); /* v0=H*v1+v0, ret=<v1|H*v1> faster if v0=r1*H*v1+r2*v0 ???  */
       a1=ta[i-1]; b1=-tb[i-1]; rb=-1.0/b1; rc=tc[i][0];
       add_av(v0,-a1,v1);            /* v0+=-a1*v1 [<v0|v0>] */
       mltply_av(rb,v0);             /* v0*=-1./b1 */
       mltply_av(b1,v1);             /* v1*=    b1 */
       for (k=0;k<nev;k++)
         add_av(ev[k],tc[i][k],v0);  /* ev[k]+=rc[k]*v0; [nevk=<evk|evk>] */
       vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
    }
    for(k=0;k<nev;k++) normvec(ev[k]); /* needed? */
    if (!mpi_i) {
      printf("\n v100.t= %s i=%3d", sclock(((itime()-l1)*100+itr/2)/itr), itr);
      if (vvv) printf(" t+= %s",sclock(~0l));
    }
    return;
}
/* ------------------------------------------------------------- */

/* --- all what be needed for pthreads of diag2 --- */
typedef struct sp_diag2 {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  thbuf *hbuf;
  int *ic;
  double vv,e1,vhv,vv2,vhv2,krit;
} tp_diag2;

/*
    void-function for pthreads (one vector iteration)

    at start vhv is estimated
    store vhv_old=vhv, calculate vhv_last and vhv=vhv_new
    corrected vhv = (vhv_new-vhv_old)+(vhv_new-vhv_last) ???
    2013-01-13 bad results sq40+0=20=OK,38+2=12.10565872(n1=11)
          39+1=16,18(sym_k=-4 10 -1000,n1=2)  ./m_square 6 2 -2 6 1
          34+6=-2.11313498     OK 2013-01-17
          32+8=-8.22686823     OK 2013-01-17 a2.i56=16m vs. a0(SH=0.33)*i55*2
                                             a2.i49=8.6m*2       cplx!
                                             a2.i56=16m(SH=0.33*i55*2) dbl! 
         1*16comp2.a2.i42=3.2m*16 i44=6.4m*8(SH=0.13) i49=52m*1(SH=1.02*i55*2)
               t(AMD)=3*t(i686)! (n1=482k)*(5+4)  scplx!
          30+10=-13.57780123   OK 2013-01-18 a2.i39=33m(SH=0.80m) flt n1=5.3e6
             a2 is 2.3*faster (unfrustrated)
      ./m_tilings 8 6 0 2 2 # k=0+ n1=32e6 -15.78155511 i130*(sh=10h)
         z=6 12+24 -12.51761327 -12.47250885 n1=26e6  mem=rs90MB/vs1.3GB
                   -12.51761248 flt a2 drift zu -12.51761344
          1*16comp2.a0.i120=(SH=1.17m*i120*2*16)=281m*16 a2.i669=964m*16
          iteration 30-60 sehr linear? E0+.06=-12.51...E0+0.04 aber flacher
          relaxationsprozess analog Abkuehlung? 0.12*exp(-0.018*x)+E0 (logy)
         z=8 10+26  -9.3335577   -9.2716939  n1=1.8e6 mem=real24MB,virt1.3G
          1*16comp2.a0.i100=(SH=0.22*i100*2*16)=44m*16 a2.i199=54m*16
                    -9.333558024 
             a2 is ...*slower (strong frustrated)
    ToDo: +ahead +checkpt +mpi
    ToDo: mpi_n>1 && nahead>1  failed 2013-02-07 !!!!???
 */
void diag2_thread( void *ptr ){
    tp_diag2 *p=(tp_diag2 *)ptr;
    tvec *v0; scplx vnew,vold;
    cplx b,c,nv,nhv,b1,r;
    double e1,r1,nhn,vhv,vv,krit=0;
    int ic,blk; ULong nn,ii;
    blk=p->hbuf->block;
    e1 =p->e1;
    vhv=p->vhv;
    vv =p->vv;
    ic =*(p->ic);
    v0 =p->hbuf->v0;  /* |v> */
    // to ensure mpi balance, loop over len_full_block and skip dummy elements
    // old: for (ii=0,n=b_ofs[blk];n<b_ofs[blk+1];n++,ii++) // no MPI balance
    for (ii=0,nn=b_ofs[blk];nn<b_ofs[blk]+len_full_block;nn++,ii++) // mpi
    {    /* like MC and fixed walk path? */
       /* n is the index within mpi-thread (v0[nn]), ii is ahead-idx ??? */
       vold=v0[nn];                         /* float, double, cplx ... */
       nv=M_f2d(vold);                     /* <nv> nn-th component of |v> */
       /* dont use disk, because handling of async IO is difficult v2.26 */
       /* nhv: <nHv> nn-th component of H|v>, <nHv> = sum_j H_ij*|v_j> */
       /* nhn: <n|Hn> diagonal-H of |n> (including <nHn>???) |n>H<n| ??? */
       // ToDo: MPI-safe??? slow, deadlock if b_ofs[blk+1] differ! 2013-02
       // ToDo: only call every NUM_AHEAD_LINES;  2013-02-06
       // old: hamilton_nhv(n-b_ofs[blk],/*ahead=*/1,p->hbuf,/*mode=*/0);
       // Caution! MPI_Send/recv inside hamilton_nhv !!!
       //  for nahead>1 old v0 where used, which could be updated for nahead=1
#define NAHEAD 1 /* NUM_AHEAD_LINES */
       if (ii%NAHEAD == 0) // speedup MPI by blocking send/recv
         hamilton_nhv( nn-b_ofs[blk], NAHEAD, p->hbuf, /*mode=*/0);
       /* returns not (jj,rr) as usual, but nhv list + nhn */
       if (nn>=b_ofs[blk+1]) continue; // nhv ignored, only mpi balance
       // old: nhv=  p->hbuf->el[0].rr; /* 2013-01 ok for ahead=1 */
       // old: nhn=  p->hbuf->nhn;
       // nhv=  p->hbuf->el[0].rr; /* 2013-01 ok for ahead=1 incl. nhn !? */
       nhv=  p->hbuf->nhv[ii%NAHEAD]; /* 2013-02-07 sum H[n,i]*|v0i> */
       nhn=  p->hbuf->nhn[ii%NAHEAD]; /* 2013-02-07 H[n,n] */
       /* nhn,nhv= 33,6.84..1.957217 32,11.48..-3.579710  32,2.1..-2.999792 (ud=2+38) */
       /* nhn,nhv= 17,6.84..1.957217 16,11.48..-3.579710  16,2.1..-2.999792 (ud=2+38) */
       if (n1<100 && mpi_i<=4) // mpiexec --tag-output
         printf("DBG diag2 ii,n nhn,nhv,nv %2d %2d %7.4f %8.4f %8.4f\n",
                (int)ii, (int)nn, nhn, D(nhv), D(nv));
       /* nhn to small for sym until v2.25 */
       /* ToDo: can iteration be splitted into sub-H-blocks? H1,H2,H3 etc?
        *       if not, we have to write paramindex per nonzero element */ 
       if(ic==0){ vhv+=D(Conj(nv)*nhv); continue; } /* only calculate vhv_last */
       /* ----------------------- matrix element of S^(-1)H*(1-nv*nv)  */
       b=vv*nhv-vhv*nv; /* a=vv*nhn-nv*nhv; */ if(fabs(b)<1e-30)continue;
       c=   nhv-nhn*nv; /* d=   vhv-nhv*nv; */
       /* <b|c>==vv*<nhv|nhv>+vhv*nhn*<nv|nv>-vv*nhn*<nhv|nv>-vhv*<nv|nhv> ??? */
       r1=vhv-vv*nhn;   /* d-a */
       if(fabs(b*c)<1e-12){continue;}  /* no changes in v0 */
       /* ToDo: sqrt is slow on some processors! AMD,MIPS? */
       r =r1-sqrt(r1*r1+4*b*c);	/* complex??? ew r1^2-4bc=(a+d)^2-ad-bc=(p/2)^2-q */
       b1=2*b/r;                /* correction b1 is small??? */
       /* b1 may be -2.7 which make nv from 0.23 to -2.47 (much to big!)
        *   resulting in inaccuracy, to small vhv which causes divergence */
       vnew=vold+M_d2f(b1);     /* rounding error ??? vold==nv */
       b1=M_f2d(vnew-vold);     /* set b1 according to rounding error */
       /* updating v0[nn] will invalidate older nHv and its derivates */
       v0[nn]=vnew; /* ToDo2013 do this nahead-block-wise buffered!? */
       krit+=Norm(b1);	/* Norm(old.v-new.v), better use max(fabs(b1))??? */
       /* vv =<v+b|v+b>  =<v|v>  +2<b|nv>  +<b|b> */
       /* vhv=<v+b|H|v+b>=<v|H|v>+2<bn|Hv>+<b|H|b> */
       /* not ok! complex ?????????, nur block-lokale Naeherungen */
       vv  +=D(2*Conj(b1)*nv )+Norm(b1);     /* correct besides matherrors */ 
       vhv +=D(2*Conj(b1)*nhv)+Norm(b1)*nhn;	/* new EW allways lower! */
       /* bad startvec can lead to overflows or big_summed_error */
       if( vv<0.5 || vv>1.6 || fabs(vhv/vv-e1)>4 ){
         printf(" warning: mpi=%d dE=%e vv=%f b1=%f\n",
           mpi_i, fabs(vhv/vv-e1), vv, D(b1));
         /* *(p->ic)=-1; */ 
         // ToDo: MPI balance? better use dummy continue? like ic=0?
         if (mpi_n==1) break; else ic=0; /* no further big changes! */
       }
    } /* for(nn,ii) */
    p->krit=krit; /* sum changes */
    p->vhv2=vhv;  /* updated vhv */
    p->vv2 =vv;   /* updated norm */
}

/* --- --- get smallest eigenvalues using special-Lanczos ----------
 *   see 2x2.tex for explanations 
 * --- Inp: iv,v0=startvector  Out: itr,ew,v0=eigenvector  ---------- */
int lanz2(tvec *v0){
    double r,e1,vhv,vv,dvhv,dvv,eps=1e-9,krit;
    int itr,ic,blk,j; ULong /* i, */ ti;
    tp_diag2 pv[B_NUM],*p; /* slow if 2 blocks fit to one cache line, ToDo */
    
    if (mpi_n>1 && !mpi_i) { E0("WARNING: lanz2 not ready for MPI"); }
    for (ic=0;ic<pt_n;ic++) pv[ic].hbuf=alloc_hbuf(); /* buffer space */
#if (VecType & 3) == 0
    eps=1e-8;  		/* float JS-Apr99 */
#endif        
#if (VecType & 8)
    eps=1e-4;  		/* float16 JS-Feb02 find more elegant way! */
#endif        
    if (!mpi_i) printf(" lanz2: eps=%5.0e\n",eps);
    vhv=0.0; e1=13.5; vv=1.0; ti=itime();
    for(itr=1,ic=0;itr<4*MAX_ITR;itr++,ic=(ic+1)&31){
       /* ToDo: reset ic on external signal sig1==1 ??? */
       krit=0;
       if (itr<8 && (itr&1)==0) ic=0; /* 2013-02-09 testwise fr mpi */ 
       if (ic==0) { vhv=0; krit=999; }
       for(blk=0;blk<pt_n;blk++){ p=&(pv[blk]);
         p->hbuf->v0   =v0;
         p->hbuf->block=blk;
         p->vv         =vv;
         p->vhv        =vhv;
         p->e1         =e1;
         p->ic         =&ic;
         parallel_start(blk, diag2_thread, p);
       }
       dvhv=dvv=0; /* 2013-02 MPI adaption, summ all changes */
       for(blk=0;blk<pt_n;blk++){ p=&(pv[blk]);
         parallel_wait(blk);
         /* collect all block-local changes */
         dvv  += p->vv2  - p->vv;
         dvhv += p->vhv2 - p->vhv;
         krit+= p->krit;
       }
#ifdef CONFIG_MPI  /* 2013-02-06 JS add MPI code */
       { double add[3], sum[3];
         add[0]=dvv; add[1]=dvhv; add[2]=krit;
         MPI_Allreduce(add, sum, 3, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
         dvv=sum[0]; dvhv=sum[1]; krit=sum[2];
       }
#endif
       vv+=dvv; vhv+=dvhv; /* sum block-local changes, not exact for ic!=0 */
       r=1/sqrt(vv);  mltply_av(r,v0); /* v0*=r */
       krit*=fabs(r); krit=sqrt(krit/(double)n1);
       e1=vhv/=vv; vv=1.0;		/* normieren */
       /* if(!(itr%5)) */
       { if (!mpi_i)
           printf("%3d %20.15f %5.0e %4s %2d %6.0e\n",
                  itr,D(e1),(double)krit,sclock(~0l),ic,1-r);
         if (krit<eps) break;
         if (krit>1.1 && ic>0) ic=-1; // 2013-02-09 recompute correct vhv
       }
       /* here the user can stop iterations (ex: slow convergence) and 
        * continue with saving vector and do the next steps,
        * usefull if computation must be stopped and proceeded later
        * since v2.26
        */
       if (sig1 == 2) { /* signal 2*USR2+1*USR1, stop and save vector */
         sig1=0; /* reset signal */
         printf("signal 2 received, stop iteration\n");
         break;
       }
       if(itime()-ti>60*60*12l){ ti=itime(); /* all 12h -> save v0 */
         /* savevec(v0,"d1.vec",-1,-1); # better msync? */ }
    }
    for (j=0;j<NEW;j++) ew[j]=D(e1);
    for (ic=0;ic<pt_n;ic++) free_hbuf(pv[ic].hbuf);
    if (vvv && !mpi_i) printf(" t+= %s",sclock(~0l));
    return(itr);
}

#ifndef S_double     /* JS 2011-10-28 */
#define S_double sizeof(double)
#endif
/* --------------------------  jacobi ----------------------------------- */
/*
 * jacobi does _not_ work for complex matrizes (in the present form)!
 * but for complex (hermitian) matrizes of type: A+iB
 *   use [[ A,-B], [ B, A]] as virtual matrix,
 *   in that case all eigenvalues got twice.
 *   and computing time is 9.2 times of the real time version
 *   but what about eigenvectors?
 * Driver for routine jacobi for symmetric real matrices
 * - what about parallelization? (BFG-package by I.J.Bush?)
 * sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2 (compare to MPI)
 * A: n1*n1 (complex: 2*n1*n1) (ToDo)
 * D: n1    (complex: 2*n1)
 * V: n1*n1 (complex: (2*n1)^2)
 * ToDo: Distributed Shared Mem Emulation, matrix stripes per node
 *   - transfer complete stripe j0...j0+n1/nodes of node j to node i
 *     on first touch of new j stripe (j-buffer)
 *   - transfer diag stripe + column j of stripe i to all nodes (diag buffer)
 *   - needs 2*n1^2*double/nodes memory per node (doubled for buffer)
 *     better 2 buffers for nonblocking overlapping MPI_ISend/IRecv?
 *   - give it a speedup?
 * - http://en.wikipedia.org/wiki/Jacobi_eigenvalue_algorithm
The Jacobi Method has been generalized to complex Hermitian matrices,
general nonsymmetric real and complex matrices as well as block matrices.
The Jacobi Method is also well suited for parallelism.
http://www.springerlink.com/content/m33288h888k6wg65/ G.M.Shroff1990
 unitary trafo ((cos(x),-exp(I*p)*sin(x)),(exp(-I*p)*sin(x),cos(x))
 with dpq=aqq-app, dmax=dpq +/- sqrt(spq^2+4apq*aqp)
 and tan(x)=2exp(i*p)aqp/dmax and p chosen to make tan(x) real
 * - http://www.feynarts.de/diag/   (Thomas Hahn)
     http://www.feynarts.de/diag/src/HEigensystem.c (hermitian)
     http://www.feynarts.de/diag/manual.pdf
     compact code
 * - matrix muss diagonal dominant sein (damit startwert gut (klein) ist)
 *   (youtube: JoernLoviscach 07.03 Jacobi-Verfahren)
 * --------------------------  jacobi ----------------------------------- */
//#define ROTAT2(a,i,j,k,l) g=a[i*n+j];h=a[k*n+l];
//    a[i*n+j]=g-s*(h+g*tau);a[k*n+l]=h+s*(g-h*tau);
#define ROTAT2(a,i,j,k,l) g=a[i*n+j];h=a[k*n+l];\
 if(g==0 && h==0)continue;a[i*n+j]=g-s*(h+g*tau);a[k*n+l]=h+s*(g-h*tau);
void jacobi2(double a[], int n, double d[], double v[]){
    int j,iq,ip,i,nrot;
    double theta,tau,t,s,h,g,c,*b,*z,tresh,sm;
    long l2, l1=0;
    if(!(b=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    if(!(z=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    if (v) /* JS 2011-10-28 v==0: degraded eigenvalues only mode */
    for(ip=0;ip<n;ip++){for(iq=0;iq<n;iq++)v[ip*n+iq]=0.0;  v[ip*n+ip]=1.0;}
    for(ip=0;ip<n;ip++){b[ip]=d[ip]=a[ip*n+ip];z[ip]=0.0;}
    nrot=0;
    if (vvv) printf(" start\n"); l1=itime();
    for (i=1;i<=50;i++){
      l2=itime();
      /* measure time 2011-10-28, we write n*n but (n*n-n)/2 is precise */
      sm=0.0; /* sum over lower left triangle of the matrix a (if zero?) */
      for(ip=0;ip<n-1;ip++)for(iq=ip+1;iq<n;iq++)sm+=fabs(a[ip*n+iq]);
      if (vvv && i>=1 && i<20 && l2-l1>9)
        printf(" %2d JACOBI-it sm= %.2e rots/(n*n)= %9.2f %6.2fm\n",
               i, sm, nrot*2./(n*n-n) /* half matrix */, (l2-l1)/60.);
      if(sm==0.0){ break; } /* finished */
      if (i < 4)tresh=0.2*sm/(n*n);
	else    tresh=0.0;
//      for (ip=0;ip<n-1;ip++) {      /* one     column  ip */
//	for (iq=ip+1;iq<n;iq++) {   /* another column  iq   9m */
      for (iq=1;iq<n;iq++) {      /* one     column  iq */
	for (ip=0;ip<iq;ip++) {   /* another column  ip faster!?  8m */
	 g=100.0*fabs(a[ip*n+iq]);  /* cross point element (cplx) */
	 if (i>4 && (fabs(d[ip])+g)==fabs(d[ip]) /* g=0 or -2d(?) */
		 && (fabs(d[iq])+g)==fabs(d[iq])) a[ip*n+iq]=0.0;
	 else if (fabs(a[ip*n+iq]) > tresh) {
	  h=d[iq]-d[ip];  /* y=(dq-dp)/2 p=apq t=|y|+sqrt(p^2+y^2) */
	  if ((fabs(h)+g)==fabs(h)) t=(a[ip*n+iq])/h; 
	  else { theta=0.5*h/(a[ip*n+iq]);
		 t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
		 if (theta < 0.0) t = -t;}
	  c=1.0/sqrt(1+t*t); s=t*c; /* c=cos=t/q(pp+tt) s=sin=p/(pp+tt)  c*c+s*s=1 */
	  tau=s/(1.0+c); h=t*a[ip*n+iq];
	  z[ip] -= h; z[iq] += h;
	  d[ip] -= h; d[iq] += h; a[ip*n+iq]=0.0;
	  for (j=0;j<=ip-1;j++)    { ROTAT2(a,j,ip,j,iq)}
	  for (j=ip+1;j<=iq-1;j++) { ROTAT2(a,ip,j,j,iq)}
	  for (j=iq+1;j<n;j++)     { ROTAT2(a,ip,j,iq,j)}
          if (v) /* JS 2011-10-28 v==0: degraded eigenvalues only mode */
	  for (j=0;j<n;j++)        { ROTAT2(v,j,ip,j,iq)}
	  ++nrot;
	 }
	}
      }
      for (ip=0;ip<n;ip++){ b[ip]+=z[ip]; d[ip]=b[ip]; z[ip]=0.0;}
    } 
    if(i>50) E0("Too many iterations in routine jacobi");
    free(z); free(b);
    printf(" rotations: %3d",nrot);
    return;
}
#undef ROTAT2
/* ----------------------------------- end jacobi ---------------------- */
/* http://www.feynarts.de/diag/ Hermitian Matrix Diag */
/*
   HEigensystem diagonalizes a Hermitian n-by-n matrix.
   Input: n, A = n-by-n matrix, Hermitian
   (only the upper triangle of A needs to be filled).
   Output: d = vector of eigenvalues, U = transformation matrix
   these fulfill diag(d) = U A U^+ = U A U^-1 with U unitary.
   Added cplx-version 
     by JS 2011-12-10 to remove [[A,-B][B,A]] trick for real matrix
 * 2011-12-13 LC18 q00 n1=765 -15.17070224  n1**2=585K cache=2MB
 *  jacobi nrots=5603160 82s            (iq,ip changed 76s)
 *  jacobi nrots=5603160 76s skip-zeros (iq,ip changed 69s)
 * cjacobi nrots=2706125 31s            (p,q changed 38s)
 * cjacobi nrots=2706125 34s skip-zeros 
 * 2011-12-13 LC18 q0x n1=1367 -15.17070224  n1**2=1.87M cache=2MB
 *  jacobi nrots=19182472               (iq,ip changed 9m)
 *  jacobi nrots=19162688 9m skip-zeros (iq,ip changed 8m)
 * cjacobi nrots=9171575 204s            (p,q changed 4.27m)
 * cjacobi nrots=9171575 225s skip-zeros (p,q changed 4.53m)
 */            
// #include "defs.h"
// ToDo: OpenMP ??? test omp sleep end, new omp sleep end (+ps uxm)
// #include <omp.h>  /* OpenMP for gcc add -fopenmp to cc+linker */
// http://gcc.gnu.org/onlinedocs/libgomp/Environment-Variables.html
// OMP_NUM_THREADS=4 ./a.out
static inline double sq(const double x) { return x*x; }
/* complex jacobi-like diagonalization */
void cjacobi(const unsigned long n, scplx *A, double *d, scplx *U)
{  unsigned long p, q, nrots=0, nrout=1024; int nsweeps;
   const double red = .04/(n*n*n*n);
   double (*ev)[2], wt2, wt1;
   
   ev=(double (*)[2])malloc(n*2*sizeof(double));
   if (!ev) { E0("cjacobi.malloc_ev failed");return; }
   if (((n-1)*n+(n-1))!=(n-1)*1.*n+(n-1)) E0("cjacobi.n overflow");
   printf(" start cjacobi\n");
//   printf(" start cjacobi OMP=%d\n", omp_get_thread_num());
   for (p=0; p<n; p++) {
     ev[p][0]= 0;
     ev[p][1]= d[p]= Re(A[p*n+p]);
     if (!U) continue;
     for (q=0; q<n; q++) U[p*n+q]= 0.0; /* unity matrix */
     U[p*n+p]= 1.0;
   }
   wt1=fstime();
   printf(" begin iterations n1= %lu\n", n1);
   for( nsweeps=1; nsweeps<=50; nsweeps++ ) {
     double thresh=0;
     /* check for near zero offdiag */
//     for (q=1;q<n;q++) for(p=0;p<q;p++)    // 31s
     for (p=0;p<n-1;p++) for(q=p+1;q<n;q++)  // 31s cache friendly
      thresh += Norm(A[p*n+q]);
     wt2=fstime();
     if(wt2==wt1) wt2=wt1+1;
     /* ToDo: count x to 0 and 0 to x rotations for stats too? */
     printf("DBG L%d: it= %2d sm= %.2e nrots= %9lu t= %6.2fm rps= %.2e\n",
        __LINE__, nsweeps, thresh, nrots, (wt2-wt1)/60., nrots*1.0*n/(wt2-wt1));
     if ( !(thresh > 0x1p-103) ) break; // 2^-103 13Bytes=104bit
     thresh = (nsweeps < 4) ? thresh*red : 0;
     /* n(n-1) rotations 0,1 0,2 1,2 0,3 1,3 2,3 0,4 1,4 2,4 ... */
     /* 0,2 and 1,2 conflict and could not be done in parallel */
     for ( q=1; q<n; q++ )
       for ( p=0; p<q; p++ ) // 31s blocking: pb=0,pb+=pblk + for (p=pb..pb+pblk)
//     for ( p=0; p<n-1; p++ ) for ( q=p+1; q<n; q++ ) // 38s cache friendly? 
     {
        const cplx Apq = A[p*n+q];
        const double off = Norm(Apq);
        if( nsweeps > 4 && off < 0x1p-103*(sq(ev[p][1]) + sq(ev[q][1])) )
          A[p*n+q] = 0;
        else if( off > thresh ) {
          double delta, t, s, invc;
          unsigned long j;
          
          nrots++; /* JS-2011-12-12 for stats */
          if (nrots==nrout) { /* for more output on long run benchmarks */
            wt2=fstime(); if(wt2==wt1) wt2=wt1+1;
            if (wt2-wt1>60) {
     printf("DBG L%d:  qp= %6lu %6lu  nrots= %9lu t= %6.2fm rps= %.2e\n",
        __LINE__, q, p, nrots, (wt2-wt1)/60., nrots*1.0*n/(wt2-wt1));
            }
            nrout<<=1;
          }
          t = .5*(ev[p][1] - ev[q][1]);
          t = 1/(t + copysign(sqrt(t*t + off), t));
          delta = off*t;
          ev[p][1] = d[p] + (ev[p][0] += delta);
          ev[q][1] = d[q] + (ev[q][0] -= delta);
          invc = sqrt(delta*t + 1);
          s = t/invc;
          t = delta/(invc + 1);
// OMP_NUM_THREADS=4
//#pragma omp parallel shared(A) private(j)
{
//#pragma omp sections nowait
{
//#pragma omp section
{
          for( j = 0; j < p; j++ ) {
            const cplx x = A[j*n+p]; // cache unfriendly +n +n +n ...
            const cplx y = A[j*n+q];
            A[j*n+p] = x + s*(Conj(Apq)*y - t*x);
            A[j*n+q] = y - s*(Apq*x + t*y);
          }
          for( j = p + 1; j < q; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[j*n+q];
            A[p*n+j] = x + s*(Apq*Conj(y) - t*x);
            A[j*n+q] = y - s*(Apq*Conj(x) + t*y);
          }
          for( j = q + 1; j < n; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[q*n+j];
            A[p*n+j] = x + s*(Apq*y - t*x);
            A[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }

          A[p*n+q] = 0;
}
//#pragma omp section
          if (U)
          for( j = 0; j < n; ++j ) { // cache friendly +1 +1 +1 ...
            const cplx x = U[p*n+j];
            const cplx y = U[q*n+j];
//            if (x==0 && y==0) continue; /* speedup!? */
            U[p*n+j] = x + s*(Apq*y - t*x);
            U[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }
} /* pragma parallel sections */
} /* pragma parallel */
        }
      } // p,q loop, upper right triangle
    for( p = 0; p < n; p++ ) {
      ev[p][0] = 0;
      d[p] = ev[p][1];
    }
  } // sweeps
  printf(" sweeps+rotations: %3d %9lu", nsweeps, nrots);
/* sort the eigenvalues */

#if 0
  for( p = 0; p < n - 1; p++ ) {
    int j = p, sort = +1 /* +1 or -1 for low-high, high-low */;
    double t = d[p];
    for( q = p + 1; q < n; q++ )
      if( sort*(t - d[q]) > 0 ) t = d[j = q];
    if( j == p ) continue;
    d[j] = d[p];
    d[p] = t;
    if (U)
    for( q = 0; q < n; ++q ) {
      const cplx x = U[p*n+q];
      U[p*n+q] = U[j*n+q];
      U[j*n+q] = x;
    }
  }  
#endif
}
              
/* parallel complex jacobi-like diagonalization */
/*
 * CPU-cache-misses? (n*nrots)/time vs. 4nn (quantum 2E5450-3.0GHz-6MBcache)
 *  below 1 It, rts starts 30% faster (cached!)
 * ziel=142e3 81GBflt!? (AH-tUcubeoctahedron),  70e3 19GBflt (JR)
 *  t=n**2*It*n/speed=.../6e7 t1/t2=(n1/n2)**3 *10**3 = 
 * LC24,16+8q00 n= 15581  16777216n/47min= 9.27e7--/s 4nn=971MB (skip-zero?)
 * LC24,17+7q0x n= 14421 318289605n/985m=  7.77e7/s,5It  flt
 * Jxy=0.01
 * LC24,17+7q0x n= 14421 505182738n/1372m= 8.85e7/s,6It++ dbl bis_q=1543 2.40e8/s dbl
 * LC24,17+7q00 n=  7293  12858223n/19.6m= 7.95e7/s 4nn=213MB flt
 * LC24,17+7q00 n=  7293  24448828n/31.7m= 9.37e7/s,2It  -1.04082821 dbl 325m=5h bis_q=2133 1.86e8/s dbl (UpperRight-tri-matrix)
 * LC24,18+6q0x n=  5620 152997208n/163m = 8.77e7/s,8It  -0.04992962 flt
 * LC24,18+6q0x n=  5620 149560852n/137m = 1.02e8/s,16It -0.04992956 dbl bis_q=2174 1.72e8/s dbl
 * LC24,18+6q00 n=  2920  38019960n/19.8m= 9.34e7/s,15It -0.04992960 flt 4nn=34MB
 * LC24,19+5q0x n=  1771  13735453n/3.89m= 1.04e8/s,14It +0.96660992 flt
 * LC24,19+5q00 n=   913   3334835n/0.45m= 1.12e8/s,12It +0.96660994 flt 4nn=3.3MB
 *                         3336389n/0.28m= 1.82e8/s,12It +0.96660994 dbl
 *    NB=4                                 1.13e8/s      +0.96661053 flt
 *    NB=32                                1.11e8/s      +0.96660992 flt
 *    NB=128               3353230n/0.50m  1.11e8/s      +0.96662083 flt
 *    NB=128               3333292n/0.30m  1.70e8/s      +0.96662334 dbl ??
 * LC24,20+4q0x n=   446    787192n/0.05m= 1.21e8/s,13It +1.96241698 flt
 * LC24,20+4q00 n=   256    234842n/0.01m= 1.20e8/s,12It +1.96241699 flt
 *                          234202n/0.01m= 1.85e8/s,11It +1.96241699 dbl
 * LC24,21+3q0x n=    85     24229n/0.00m= 1.13e8/s,11It +2.98544364 +5.00005025 flt
 * LC24,21+3q00 n=    48      6520n/0.00m= 1.01e8/s,9It  +2.98544367 flt
 * LC24,22+2q00 n=    12       304n/0.00s= 1.94e7/s,7It  +3.98020321 +5.00010000 flt
 *                                         2.29e7/s,7It  +3.98020321 dbl
 */
#if 0 // in progress (testing/debugging)
void pcjacobi(const unsigned long n, scplx *A, double *d, scplx *U)
{  unsigned long p, q, nrots=0, nrout=1024; int nsweeps;
   const double red = .04/(n*n*n*n);
   double (*ev)[2], wt2, wt1;
   
   ev=(double (*)[2])malloc(n*2*sizeof(double));
   if (!ev) { E0("cjacobi.malloc_ev failed");return; }
   if (((n-1)*n+(n-1))!=(n-1)*1.*n+(n-1)) E0("cjacobi.n overflow");
   printf(" start pcjacobi\n");
//   printf(" start cjacobi OMP=%d\n", omp_get_thread_num());
   for (p=0; p<n; p++) {
     ev[p][0]= 0;
     ev[p][1]= d[p]= Re(A[p*n+p]);
     if (!U) continue;
     for (q=0; q<n; q++) U[p*n+q]= 0.0; /* unity matrix */
     U[p*n+p]= 1.0;
   }
   wt1=fstime();
   printf(" begin iterations n1= %lu\n", n1);
   for( nsweeps=1; nsweeps<=50; nsweeps++ ) {
     double thresh=0;
#define NB 4  /* blocking buffer for better cache use 2011-11-16 */
     double bs[NB], bt[NB]; unsigned long bp;
     cplx bApq[NB];
     /* check for near zero offdiag */
//     for (q=1;q<n;q++) for(p=0;p<q;p++)    // 31s
     for (p=0;p<n-1;p++) for(q=p+1;q<n;q++)  // 31s cache friendly
      thresh += Norm(A[p*n+q]);
     wt2=fstime(); if(wt2==wt1) wt2=wt1+1;
     /* ToDo: count x to 0 and 0 to x rotations for stats too? */
     printf("DBG L%d: it= %2d sm= %.2e nrots= %9lu t= %6.2fm rps= %.2e\n",
        __LINE__, nsweeps, thresh, nrots, (wt2-wt1)/60., nrots*1.0*n/(wt2-wt1));
     if ( !(thresh > 0x1p-103) ) break; // 2^-103 13Bytes=104bit
     thresh = (nsweeps < 4) ? thresh*red : 0;
     /* n(n-1) rotations 0,1 0,2 1,2 0,3 1,3 2,3 0,4 1,4 2,4 ... */
     /* 0,2 and 1,2 conflict and could not be done in parallel */
     for ( q=1; q<n; q++ )  // upper right triangle
       for ( bp=0; bp<q; bp+=NB ) // 31s blocking: pb=0,pb+=pblk + for (p=pb..pb+pblk)
//         for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
//     for ( p=0; p<n-1; p++ ) for ( q=p+1; q<n; q++ ) // 38s cache friendly? 
     {
          unsigned long j;
          if (nrots>nrout) { /* for more output on long run benchmarks */
            wt2=fstime(); if(wt2==wt1) wt2=wt1+1;
            if (wt2-wt1>60) {
     printf("DBG L%d:  qp= %6lu %6lu  nrots= %9lu t= %6.2fm rps= %.2e\n",
        __LINE__, q, p, nrots, (wt2-wt1)/60., nrots*1.0*n/(wt2-wt1));
            }
            nrout<<=1;
          }
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
  {
        const cplx Apq = A[p*n+q];     
        const double off = Norm(Apq);  // p<q
        if( nsweeps > 4 && off < 0x1p-103*(sq(ev[p][1]) + sq(ev[q][1])) )
          A[p*n+q] = 0;
        else if( off > thresh ) {
          double delta, t, s, invc;
          
          nrots++; /* JS-2011-12-12 for stats */
          t = .5*(ev[p][1] - ev[q][1]); // p<q  diag
          t = 1/(t + copysign(sqrt(t*t + off), t));
          delta = off*t;
          ev[p][1] = d[p] + (ev[p][0] += delta);
          ev[q][1] = d[q] + (ev[q][0] -= delta);
          invc = sqrt(delta*t + 1);
          s = t/invc;
          t = delta/(invc + 1);
  bApq[p-bp]=Apq;
  bs[p-bp]=s;
  bt[p-bp]=t;
  // compute overlapping elements (exclude below!)
if(0) // seperate outer block and shift below
           for( j = 0; j < bp; j++ )  { // bp<=p<q downwards
            const cplx x = A[j*n+p]; // cache unfriendly +n +n +n ...
            const cplx y = A[j*n+q];
            A[j*n+p] = x + s*(Conj(Apq)*y - t*x);
            A[j*n+q] = y - s*(Apq*x + t*y);
          }
           for( j = bp; j < p; j++ )  { // p<q downwards
            const cplx x = A[j*n+p]; // cache unfriendly +n +n +n ...
            const cplx y = A[j*n+q];
            A[j*n+p] = x + s*(Conj(Apq)*y - t*x);
            A[j*n+q] = y - s*(Apq*x + t*y);
          }
          for( j = p+1; j < q && j*0<bp+NB; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[j*n+q];
            A[p*n+j] = x + s*(Apq*Conj(y) - t*x);
            A[j*n+q] = y - s*(Apq*Conj(x) + t*y);
          }
          for( j = max(q+1,0*bp); j < n && j*0<bp+NB; ++j ) { // p<q
            const cplx x = A[p*n+j];
            const cplx y = A[q*n+j];
            A[p*n+j] = x + s*(Apq*y - t*x);
            A[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }
          A[p*n+q] = 0;
      } // A!=0 rot
} // p-loop
// OMP_NUM_THREADS=4
//#pragma omp parallel shared(A) private(j)
//#pragma omp sections nowait
//#pragma omp section
// ToDo: error because Apq=A[p,q] is modified within bp loop
//   do all things outside buffer-loop
if(1)
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
  if (Norm(bApq[p-bp])>thresh)
      for( j = 0; j < bp; j++ )
           {
            const cplx x = A[j*n+p]; // cache unfriendly +n +n +n ...
            const cplx y = A[j*n+q];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[j*n+p] = x + s*(Conj(Apq)*y - t*x);
            A[j*n+q] = y - s*(Apq*x + t*y);
          }
if (0) {
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
      for( j = bp+NB; j < p; j++ )
           {
            const cplx x = A[j*n+p]; // cache unfriendly +n +n +n ...
            const cplx y = A[j*n+q];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[j*n+p] = x + s*(Conj(Apq)*y - t*x);
            A[j*n+q] = y - s*(Apq*x + t*y);
          }
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
          for( j = p+1; j < q && j<bp; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[j*n+q];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[p*n+j] = x + s*(Apq*Conj(y) - t*x);
            A[j*n+q] = y - s*(Apq*Conj(x) + t*y);
          }
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
          for( j = max(bp+NB,p+1); j < q; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[j*n+q];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[p*n+j] = x + s*(Apq*Conj(y) - t*x);
            A[j*n+q] = y - s*(Apq*Conj(x) + t*y);
          }
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
          for( j = q+1; j < n && j<bp; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[q*n+j];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[p*n+j] = x + s*(Apq*y - t*x);
            A[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
          for( j = max(bp+NB,q+1); j < n; ++j ) {
            const cplx x = A[p*n+j];
            const cplx y = A[q*n+j];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
            A[p*n+j] = x + s*(Apq*y - t*x);
            A[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }
//#pragma omp section
          if (U)
for ( p=bp; p<q && p<bp+NB; p++) // ToDo: shift inside
          for( j = 0; j < n; ++j ) { // cache friendly +1 +1 +1 ...
            const cplx x = U[p*n+j];
            const cplx y = U[q*n+j];
            const double s=bs[p-bp], t=bt[p-bp]; const cplx Apq=bApq[p-bp];
//            if (x==0 && y==0) continue; /* speedup!? */
            U[p*n+j] = x + s*(Apq*y - t*x);
            U[q*n+j] = y - s*(Conj(Apq)*x + t*y);
          }
} // bp-loop

    if (nsweeps==1 && n<20) { unsigned int ip, iq;
     for ( ip=0; ip<n; ip++ ) { // upper right triangle
       for ( iq=0; iq<n; iq++ )  // 31s blocking: pb=0,pb+=pblk + for (p=pb..pb+pb
         if (iq<ip) printf(" %6s", ((iq==bp && ip==q)?"-*-":" "));   
         else { if (A[ip*n+iq]==0) printf(" %6d", 0);
                else printf(" %6.0e", D(A[ip*n+iq])); }
         printf("\n");
     } }

      } // p,q-loop upper right triangle
    for( p = 0; p < n; p++ ) {
      ev[p][0] = 0;
      d[p] = ev[p][1];
    }
  } // sweeps
  printf(" sweeps+rotations: %3d %9lu", nsweeps, nrots);
/* sort the eigenvalues */

#if 0
  for( p = 0; p < n - 1; p++ ) {
    int j = p, sort = +1 /* +1 or -1 for low-high, high-low */;
    double t = d[p];
    for( q = p + 1; q < n; q++ )
      if( sort*(t - d[q]) > 0 ) t = d[j = q];
    if( j == p ) continue;
    d[j] = d[p];
    d[p] = t;
    if (U)
    for( q = 0; q < n; ++q ) {
      const cplx x = U[p*n+q];
      U[p*n+q] = U[j*n+q];
      U[j*n+q] = x;
    }
  }  
#endif
}
#endif // if 0

              
void shellsort(double a[],int ia[],ULong n);
/* call jacobi to diagonalize tridiag for testcase
 * --------------------------  Matrizen ----------------------------------- */
void fulltridiag(unsigned int n1,double ta[], double tb[]){
    int *ix; ULong i,j;
    double *xd,xi;     /* Matrizen 1 & 2 */
    scplx *xm; /* we need only real, but HR... is more general for cplx */
    if(vvv)printf("\nfulltridiag"); xi=0;
    /* standart jacobi and LAPACK expects simple vectors, convert block form */
    /* add 1 for the case n1=1, where Lapack fails */
    xm=(scplx *)mymalloc(n1*n1*sizeof(scplx),NULL); /* Matrix_H-in */
    xd=(double *)mymalloc(   n1*S_double,NULL); /* eigenvalues-out */
    ix=   (int *)mymalloc(   n1*S_int,NULL);    /* sort-idx-out */
    for(i=0;i<n1;i++) ix[i]=(int)i; /* initialize sort-index (LAPACK) */
    if(xm!=NULL /* && xv!=NULL */ && xd!=NULL && ix!=NULL){
      if (vvv) printf(" H");
      /* generate a simple matrix, store all elements to xm[] */
      for(i=0;i<n1;i++) for(j=0;j<n1;j++){ xm[i*n1+j]=0.0; }
      for(i=0;i<n1;i++) xm[i*n1+i]=ta[i]; 
      for(i=0;i<n1-1;i++) { xm[i*n1+i+1]=tb[i]; xm[(i+1)*n1+i]=tb[i]; }
      printf(" JACOBI");
      /* jacobi2(&xm[0],(int)(n1),xd,NULL); */
      /* HEigensystem(const int n, cplx *A, double *d, cplx *U); */
      cjacobi((int)(n1),xm,xd,NULL);
      /* ToDo: if xd==0 auch ohne sort! */
      shellsort(xd,ix,n1);		/* sort xd[ix[i]] */
      printf("\n eigenvals=");
      for (j=0;j<n1;j++){ printf(" %+12.8f",xd[ix[j]]); } printf("\n");
    } /* else errormessage ToDo */ 
    myfree((void *)ix,NULL);
    myfree((void *)xm,NULL);
    myfree((void *)xd,NULL);
    return;
}
