const char *version_string="2013-11-07";
// #define CFG_OMP  /* define CFG_OMP for openmp + gcc -fopenmp ... */
// #define CFG_MPI        /* define CFG_MPI + mpicc ... */ 
// #define CFG_CPUSET     /* options --pin_ncpu=... */
/*
 measure speed of memory (best and worst case)
 also FDIR
  - Ausfaelle erkennen (fault-detection)          
  - Ausfaelle einkreisen und bestimmen (fault-isolation)
  - geeignete Korrekturen (recovery)    
 
 INST_SEQ: # single task (no parallelism)
          gcc -Wall -O2 -o memspeed memspeed.c
          # memory footprint: 
 INST_MPI: # OpenMPI-usage (parallel on distributed memory)
          mpicc -Wall -O2 -o memspeed memspeed.c
          # optional: mpicc|gcc -DCFG_OMP -fopenmp ...
          #   cpuset is a hex bit pattern for used cpu cores (p.e. 0x05,0x0f)
          #     useful to bind tasks to cores using sched_setaffinity()
          #   use cc or gcc if mpicc is not availabe
          mpirun -np <tasks> ./memspeed --log2size=<log2size> --mintime=<sec>
          # needs: memory = tasks*2^log2size Bytes
          #        time   = (1..2)*2*mintime*(log2size-8) seconds
          mpirun -np 8 --bynode ./memspeed --log2size=29 #    4GB=8*512MB
          mpirun -np 2 --bynode ./memspeed --log2size=27 #  256MB=2*128MB
          mpirun -np 2 --bynode ./memspeed --asize=384   #  768MB=2*384MB
          # runtime is about 2*(minsize...2minsize)*(log2size-7)
          ... --pin_ncpu=6 ...
 INST_OMP:# OpenMP-usage (parallel on shared memory)
          yum install libgomp
          gcc -Wall -O2 -DCFG_OMP -fopenmp -o memspeed memspeed.c
          export OMP_WAIT_POLICY="PASSIVE" # OpenMP-3.0 waits saving power 
          export OMP_NUM_THREADS=8         # OpenMP     8 threads
          OMP_PROC_BIND=TRUE # thread bounding, thread affinity
          OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./memspeed
          # memory footprint: 
 MONITOR: ps -Lf     # UID PID PPID LWP(!) C NLWP STIME TTY TIME CMD
          ps -T      # PID SPID TIME TTY CMD
          top -H
          cat /sys/devices/system/edac/mc/mc[0-3]/ce_count | xargs # ECC-Errors
          cat /sys/devices/system/edac/mc/poll_msec # 1000 ms # x86+amd
 TEST-Suites:
          kautz2009-2012 lot of ECC-Errors and types (CE,UE,panics)
          comp2-ECC-CEs+Crashs

ToDo: better read random index from array?
 - reduce code size and memory footprint (90M-180M openmpi), mpi-less?
 - check FLOPs for mem+power stress using oprofile
 - modules speed random, random_prefetch, seq, r[w], + stress + mark badmem
   MEMstress +FPU+GPU+DSK20W+DVD+USB2W+ETH+IB(=PSU)-stress? vs. Verbrauch?
   test: stresstypes vs. power on comp2
   see badRAM-linux-patch + memtest86-tests(random+mod20)
    output speed only if changed, not after 10 changes (less output) +timestamps
 - add OpenMP threads (alternative: fork/clone+pipe or pthreads)
ChangeLog:
 2013-07-02: fix problem with i(neg.long) % a2 becoming 0xfff... -gt a2
 2013-06-18: improve on 972*4GB cluster which has bitflips (kautz)
 2013-06-14: option --asize=8192 (units of log2size.default=20 fror MB)
             ini mem, to avoid page misses on 1st loop, lstream.p4
 2013-06-06: --test=2 prefetch random + --cpu_...=...
 2013-05-02: use wtime [sss.mmmuuu], not itime [sss.000000] if not MPI
             add CFG_MPI + CFG_OMP, add test8(r+w)=stresstest_mem+psu
 2008-08-22: add mpi-functionality to fill all cores of a node,
             add a random shaked pointer chain (less instructions)
  SRC-RND: do { ll=aa[ll]; } while(--k);
  SRC-STR: do { x^=*ap; ap++; } while(--k);
  ; gcc-4.3.3-x86-64 .L41 = 3 instr/loop, .L59 = 4 instr/loop
  .L52: subl    $1, %eax             ; gcc-4.1.1-i386   3 instr/loop
        movl    (%edx,%ebx,4), %ebx
        jne     .L52
  .L50: subq    $1, %rax             ; gcc-4.1.0-x86-64 3 instr/loop
        movq    (%rdx,%rbp,8), %rbp
        jne     .L50
  .L55: ... shladd r14 = r37, 3, r15  ; gcc-4.1.1-ia64  3 instr/loop
        ... ld8    r37 = [r14]
        ... br.cloop.sptk.few .L55
  L$66: s8addq  $10, $0, $10          ; cxx-6.5-alpha 19 instr/8loop
        lda     $2, -8($2)            ;  + 5 inst/loop 
        ldq     $10, ($10)
        cmple   $2, 7, $3
        s8addq  $10, $0, $10
        ldq     $10, ($10)  
        ... above 2 lines repeated 7 times (unrolled loop)
        beq     $3, L$66
                                        
 2007-04-27:  4 x86-instructions per loop (gcc-4.1.1 -O2)
              5 opteron-Instr.   per loop (gcc-4.1.0 -O2)
              4 sparcv9-Instr.   per loop (gcc-4.1.1 -O2)
             32 alpha-Instr.     per loop (cxx-6.5 -fast)
  .L15: xorl    aa(,%edx,4), %edi    ; gcc-x86-p4
        addl    %ebx, %edx
        subl    $1, %eax
        jne     .L15
  .L16: ldx     [%g3], %g1           ; gcc-sparcv9
        add     %g2, -1, %g2
        xor     %l0, %g1, %l0
        brnz,pt %g2, .L16   
  .L900000128:                       ; CC-5.3-sparcv9
      add     %o1,%i3,%o1
      xor     %l0,%o2,%l0
      add     %i2,%i4,%i2
      addcc   %o4,-1,%o4
      bne,a,pt        %xcc,.L900000128

 # P4-2.6GHz/2_maxspeed=(1.3GHz/(loop=5Instr)=260MHz)*(long=4)=1040MB/s
 # see also memspeed.gpl
 # compare with numademo 8000000000 # 8GB per run? (drop_caches! 5 vs. 8GB/s)
 #    2012comp2 memset=5-8GB/s 
 # System               Array Rand MaxMB/s Err  # new 4x->4summed
 Nehalem-2GHz-4MB       256MB   70   6338       114ns 64bit-gcc-4.3.3 72GB -O2
 Nehalem-2GHz-4MB       256MB   70   6340       114ns 64bit-gcc-4.3.3 72GB -O1
 Nehalem-2GHz-4MB       256MB   68   1850       118ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 Nehalem-2GHz-4MB    2x 256MB   66   4707       122ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 PM- 600MHz-1MB         512kB  284    393    -   14ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM- 600MHz-1MB         256MB   23    382    -  174ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM-1400MHz-1MB          64MB   31    870    -  127ns 32bit-gcc-4.1.2 2GB
 P4-2.6GHz/2-512kB      256MB   20   1090    -  197ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 P4-2.6GHz/2-512kB HT 2x256MB   26   1533    -  307ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 SunFire-880             32MB   23    220    6  346ns 64bit-gcc-4.1.1 USparcIII 750MHz
 SunFire-880             32MB   23    232    7  333ns 64bit-CC-5.3-fast-xarch=v9
 SunFire-V490            32MB   92    798   19   84ns 64bit-gcc-4.1.1 USparcIV+ 10*150MHz L2=32MB 8-way
 DOpteron275-2.2GHz      32MB   89   2130   34   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  2x  32MB  177   4129   16   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  4x  32MB  340   7156   92   91ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  6x  32MB  346   7200   98  177ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  8x  32MB  332   7248   92  189ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron885-2.6GHz      32MB   35    838   21  223ns 64bit-gcc-4.1.0 8user   hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 4HOpteron8431-2.4GHz 8x 32MB 8*79 8*1820    -  101ns 64bit-gcc-4.1.2 3user   hdparm-T=5800MB/s (summed)
 Xeon-2333Mz-4MB         32MB   69   3295       116ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     4x  32MB 4*62 4*1466       131ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x  32MB 8*55  8*734       144ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad 
 Xeon-2333Mz-4MB     1x 512MB   60   3269       133ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x 512MB 8*39  8*735       204ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad  Aug08
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-gcc-4.1.1-long
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-cxx-6.5-long
 alpha-gs1280            32MB   31    453   50      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280            32MB   46    644   22  170ns 64bit-cxx-6.5-long   (8cpus_per_rad) 8user
 alpha-gs1280            32MB   46    339  170  170ns 64bit-gcc-4.1.1-long (8cpus_per_rad) 8user
 alpha-gs1280            32MB   77   1109    9  101ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  14400 1600      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  35520  640  102ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-es45-1250MHz      32MB   38    462   56  204ns 64bit-cxx-6.5-long 4user
 altix330-IA64-1.5GHz    32MB   22    309   12  349ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 8x 32MB  168   2304   80  970ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 15x32MB  660   9300  300  176ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz    32MB   54    670   14  143ns numactl --localalloc
 altix330-IA64-1.5GHz 8x 32MB  392   4760  360  158ns numactl --localalloc
 altix330-IA64-1.5GHz 15x32MB  735  10050  300  158ns numactl --localalloc
 altix330-IA64-1.5GHz 8x512MB  418   5750       153ns min= 8*52MB/s aff+local
 altix330-IA64-1.5GHz 8x128MB  427   5748       150ns min= 8*53MB/s aff+local
 altix330-IA64-1.5GHz 2x128MB   91    709       174ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5GHz 2x128MB  107   1438       150ns min= 2*53MB/s aff0,8
 altix330-IA64-1.5GHz 2x512MB   89    710       179ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5G  16x128MB  448   6928       285ns min=16*28MB/s localalloc
 altix330-IA64-1.5G  16x128MB  700  10800       182ns min=16*44MB/s localalloc
 altix330-IA64-1.5G  16x128MB  729  11335       182ns min=16*45MB/s aff+lalloc
 altix330-IA64-1.5G  16x512MB  480   7456       264ns min=16*30MB/s localalloc
# echo 1 > /proc/sys/vm/drop_caches + numademo 1024m # ???
# 30*8quantum numademo 1024m 1node  stream=3.7GB/s     rnd=137MB/s
# 22*16comp2  numademo 1024m 4nodes stream=2.8-6.3GB/s rnd=93-125MB/s 4*16GB
# check with linkstat!
# ----------- old data (step only) --------------------
 System                 Array Min MaxMB/s Err ln2MinStep # ---- old(!) ----
 PM- 600MHz-1MB 512MB    16MB   52    364   16    12   
 PM-1400MHz-1MB 512MB    16MB   50    364    8    12   
 Xeon-3GHz-1MB  2GB     256MB   39   1800   48    16   32bit-gcc-2.96  connect model=3 stepping=4  hdparm=800MB/s  busclock=200MHz
 Xeon-3GHz-2MB  2GB     256MB   44   2700   57    20   32bit-gcc-4.0.1 bb    model=4 stepping=3  hdparm=1472MB/s
 Xeon-3GHz-2MB 12GB     256MB   39   1422   31    20   64bit-gcc-4.1.1 morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB     512MB   38   1450   31    20   32bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB    8192MB   80   1170   23    19   64bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 alpha-gs1280 128GB      32MB   36   1117   39    15   64bit-gcc-3.4.3-long (32cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   41   2485   45    15   64bit-cxx-6.5-long   (32cpus_per_rad) (32x2.5GB/s=80GB/s local) remote.maxBW=6.3GB/s local.maxBW=12.3GB/s(767MHz*8channels*2Byte) 8B/(82ns..250ns)=>32..98MB/s
 alpha-gs1280 128GB      32MB   67   1117   39    16   64bit-gcc-3.4.3-long (1cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   61   2485   47    16   64bit-cxx-6.5-long   (1cpus_per_rad) (32x2.5GB/s=80GB/s local)
 alpha-gs1280 128GB      16GB   24    572    7    23   64bit-cxx-6.5-long   (8cpus_per_rad)
 alpha-gs1280 128GB      16GB   29   1094   24    23   64bit-gcc-4.1.1-long (8cpus_per_rad)
 alpha-es45-1250MHz      16MB   87    818   20    11   cxx-int? 4/5CPU
 DOpteron275-2.2GHz-1MB  16MB   35   1500   34    13   64bit-gcc-3.4.6      hdparm-T=1900MB/s
 DOpteron885-2.6GHz-1MB 256MB   38   1120   39    15   64bit-gcc-4.0.3      hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 DOpteron885-2.6GHz-1MB 512MB   68   2300   82    15   64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 DOpteron885-2.6GHz-1MB 512MB*4 46   2220   77    15   CPU3-6! 64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 MinMax fuer 2 Jobs etc. # linkstat?
 cpu.model=generation stepping=version (same masque)
 ToDo: try __*_prefetch and improve loop (tune for generated asm-code)
 Theory vs. Samples (yum install edac-utils):
   NumCPUs*Memchannels/CPU*maxDatarate vs. DIMMs/Channel*Ranks/DIMM*Speed
 see ct23/2013 REP MOVSB different speeds, bugs (break loop, INT,VMs,...)
 Memory-Footprint: openmpi is bad! 128MB/node (tune ompi-compilation!)
 comp2: 2*AMD-6134*4Channels * max 3 DIMMs/channel = 2*42.7GB/s bei 1DIMM/ch
        2*8GB/Channel (DDR3-1333) = 10.664GB/s / channel = 85.312GB/s
          aBW=22.3GB/s? (unroll2?=26.0GB/s)  memtest4.20=3746MB/s(singleCPU)
          aBW=46.6GB/sR 2013-06
           http://en.wikipedia.org/wiki/Memory_footprint
           1MBmemusage: approx 128MB sm for openmpi
            --mca btl self,sm,tcp # 1*6M/126M ... 4*6M/191MB... 8*7M/193MB
              67MB tmp/openmpi-sessions-jschulen@comp2_0/60576/1/shared_mem_pool.comp2
            --mca btl self,tcp # 8*5M/91MB (9MB openmpi-libs)
            single-task-mpi: 2M/23MB 9MB-ompi-lib
            single-task-nompi: 1.5M/5MB 2MB-libc /proc/$PID/smaps
 quantum: 2*E5450 (8GB/s?)
          8 DDR2-667(PC-5300F)=5.3GB/s(10.6GB/s DualChannel)
          aBW=6.17GB/sRW   
 pse:   2*6core-iX5650-2.66GHz = 2*3 Speicherkanaele(DDR3-1333) max 2*32GB/s
        18*16GB-DDR3-1066MHz(a8.5GB/s) = 3 DIMMs/Channel max 51GB/s (ToDo)
          aBW=29.5GB/sRW 
 quanta1: 4*8 L1=32kB(0.5-4ns) L2=4*8*256kB(0.7-10ns) L3=16MB(1.05-33ns) mem(1.44-128ns)
          aBW1=5.5GB/s rand1=63MB/s(127ns,760MB)
  # Tasks Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]
  1 7.73e+11         7 3110.84 3.48e+09 3.48e+09   4.60 lr+w4*+16
  1 7.73e+11         7 3080.90 3.51e+09 3.51e+09   4.55 lr+w4*+16
  1 7.73e+11         7 3042.04 3.56e+09 3.56e+09   4.50 lr+w4*+16
ToDo:
 - include memtest.c
 - add pthread
 - memory-footprints (pure,mpi-???,ompi-???,stst,dyn)
  memusage(ELF64+libc6+1MB+test32) method: (ulimit -v ....;./a.out) 2013-09
  gcc4.1.2                   4.8MB       
  gcc4.1.2+static            1.9MB       
  gcc4.1.2+openmp2.5        11.2MB(1) 21.5MB(2) 31.7MB(3) 42.0MB(4) 82.9MB(8)
  gcc4.1.2+openmp2.5-static  2.0MB(1) 12.2MB(2) 22.4MB(3) 32.7MB(4) 73.7MB(8)
  openmpi2.1+gcc4.1.2        90MB(1)   n*155MB(2-8)      
  gcc4.1.2+-lpthread         6.9MB(1) 
  # RESUME2013: -fopenmp +10MB*n overhead! ompi=+154MB*n overhead
 - messe auch free-difference or max malloc (shared LIBS!)
 - einzelprogramme besser als OMPI!?
 - audiotxt verbessern arecord als stream! fuer const. mem on PC
         
*/


#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#ifdef CFG_OMP
#include <omp.h>    /* yum install libgomp; gcc -fopenmp ... */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* 2013-05 adding strncmp() */
#include <time.h>
#include <unistd.h> /* usleep */

#ifndef CFG_MPI    // for the case of MPI is not available
/* last tests: 2013-05-02 */
#include <sys/time.h>
double wtime(){ struct timeval tv;
 if(gettimeofday(&tv, NULL)){perror("timeofday");exit(2);}
 return tv.tv_sec+1e-6*tv.tv_usec; }
#define MPI_Init(a,b) 1
#define MPI_SUCCESS 1
#define MPI_Comm_size(MPI_COMM_WORLD, mpi_n)
#define MPI_Comm_rank(MPI_COMM_WORLD, mpi_i)
#define MPI_Barrier(MPI_COMM_WORLD)
#define MPI_Wtime()  wtime() /* 2013-05-02 itime replaced by wtime */
#define MPI_Finalize()
#define MPI_Abort(MPI_COMM_WORLD,xx) exit(xx)
#define MPI_Bcast(pt, len, MPI_DOUBLE, id, MPI_COMM_WORLD)
/* Reduce replacement works for MIN MAX SUM,
   Reduce to rank0, Allreduce to all ranks */
#define MPI_Reduce(snd, rec, SIZE, M_INT, M_MIN, rk, MCW) *rec=*snd
#define MPI_Allreduce(snd, rec, SIZE, M_INT, M_MIN, MCW) *rec=*snd
#else
#include <mpi.h>
#endif
#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
# include <unistd.h>  // declare sleep()+usleep
# include <sched.h>   // declare sched_setaffinity()
# ifndef _FEATURES_H  
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
# ifndef __USE_GNU                                                 
#  warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
# endif
#endif


long int *aa, alen;  // long array and size of it

void check_ring( long int a2, int vvv, int mpi_i ) {
  long int k, l;
  // check that we have still a full ring
  if (vvv && !mpi_i) printf("# chain:");
  for (l=0,k=0; k<a2; k++) {
    if (vvv && k <8 && !mpi_i) printf(" %2d", (int)l);
    if (vvv && k==8 && !mpi_i) printf(" ...");
    l=aa[l];
    if (k<a2-1 && l==0)
      printf("\n# ring stops early at %ld %.2f%%\n ...", k, k*100./a2);
  } if (vvv && !mpi_i) printf("\n");
  return;
}

int main(int argn, char *argv[]){
  double t1, t2, tmin=1.0;
  int i1, 
      omp_n=1,          /* OpenMP omp_num_threads() */
      mpi_n=1, mpi_i=0, /* MPI threads and thread number */
      pin_flag=0, /* thread to processor pinning */
      testmode=  5, /* 1=seq_read 2=pf_random 4=nonpf_random 8=stress ... */
      log2size= 20, // 2^20=1MB  set by argument1 (old, new: unit)
      log2long;     // 5 for 2^5=int32 or 6 for 2^6=int64 (long int *aa)
  long int loops, asize= 1 /* new: uses log2size as unit */,
     i, j, r00, r01, r02, r10, r11, r12, a2, k,
     a_rnd[1024], /* array of random numbers to speedup randomization */
     ll=0, xy=0; // dummy vars for the innerst loop (unique names!)
  long int memhole0=0, memhole1=0; // memory hole (bad memory) 2013-11-07
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  setvbuf(stdout, NULL, _IONBF, 0); /* dont buffer output */
  if (!mpi_i)
   printf("# memspeed %s (c) 2008-2013 GPLv2 Joerg Schulenburg ovgu.de\n",
           version_string);
  /* 2013-05 add option strings --... for more flexibility */
  for (i1=1;i1<argn;i1++) { /* check all args */
    if (strncmp(argv[i1],"--asize=",8)==0) { asize=atoi(argv[i1]+8);
      if (asize<1) asize=1; continue; } // min 1 * 2^log2size 
    if (strncmp(argv[i1],"--log2size=",11)==0) { log2size=atoi(argv[i1]+11);
      if (log2size<10) log2size=10; continue; } // min 1K 
    if (strncmp(argv[i1],"--mintime=",10)==0) { 
      tmin=atof(argv[i1]+10); if (tmin<0.0) tmin=0.0; continue; }
    if (strncmp(argv[i1],"--test=",7)==0) { 
      testmode =atoi(argv[i1]+7); continue; }
    if (strncmp(argv[i1],"--memhole=",10)==0) { int i2;
      memhole0 =atoi(argv[i1]+10);  // MB-units (1<<20)! from
      for(i2=10;argv[i1][i2]!=0
            &&  argv[i1][i2]!='-';i2++);
      if ('-'== argv[i1][i2]) memhole1= atoi(argv[i1]+i2+1); // to (exclusive)
      continue; }
    /* pin_* = processor core/thread pinning */
    if (strncmp(argv[i1],"--pin_ncpu=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_ofs=", 10)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_dist=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i1],"--pin_size=",11)==0) { pin_flag=1; continue; }
    if (!mpi_i) {
     printf(" ERROR: bad or unknown options or arguments\n"
           " use: ./memspeed --asize=16 --log2size=20 --mintime=1 --test=5\n");
     fputs("  --pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1\n",stdout);
     fputs("  --memhole=14-16   # hole of 2MB at 14MB (bad mem)\n",stdout);
    }
    usleep(100000l); /* give 100ms for output */ MPI_Abort(MPI_COMM_WORLD,1);
  }
  /* get the log2size of long int, log2long-3 is log2(Bytesize(long)) */
  for (log2long=0; 1<<log2long < 8*sizeof(long int); log2long++);
  alen = asize<<(log2size-log2long+3); /* log2(long=int32): log2long=5 */
  asize= asize<< log2size;

#if defined(CFG_CPUSET) && defined(__USE_GNU)
  if (pin_flag==1) { /* processor/thread pinning */
   // ToDo: openmp code missing
   // this processor bounding is useful for the case, that the jobsystem
   //   does not do this automatically
   // CFG_CPUSET is the bitmap for available SMP-CPUs per node
   // CFG_CPUSET=0x0005: use core 0 and 2 (Nodes with 2 Hyper-Threading CPUs)
   //   ToDo: define on runtime per taskset + sched_getaffinity ? (jobsystem?)
   // use --byslot for numbits(CFG_CPUSET)<mpi_n
   cpu_set_t cpuset; // typedef struct { ULong bits[CPU_SETSIZE / NCPUBITS]; }
   // new CPU pinning method for more than 32 cores
   struct s_pin {int ncpu;int ofs;int dist;int size;}; // new2013-05-07
   // most important value first (numcpus, then offset, then distance,...)
   int  ncores=0, ii; unsigned long cmask=0L;
   struct s_pin pin = {16,0,1,1}; // CFG_CPUSET=16,0,1,1 mask=0xffff
   // struct s_pin pin = {8,1,4,2}; // mask=0x00066
   for (i1=1;i1<argn;i1++) { /* check all args */
      if (strncmp(argv[i1],"--pin_ncpu=",11)==0) {
         pin.ncpu=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_ofs=",10)==0) {
         pin.ofs= atoi(argv[i1]+10); continue; }
      if (strncmp(argv[i1],"--pin_dist=",11)==0) {
         pin.dist=atoi(argv[i1]+11); continue; }
      if (strncmp(argv[i1],"--pin_size=",11)==0) {
         pin.size=atoi(argv[i1]+11); continue; }
   }

   if (pin.ncpu<1) {printf("ERROR: pin.ncpu<1\n");exit(1);}
   if (pin.size<1) {printf("ERROR: pin.size<1\n");exit(1);}
   if (pin.dist<0) {printf("ERROR: pin.dist<1\n");exit(1);}
   if (pin.ofs>=pin.ncpu) {printf("ERROR: pin.shift\n");exit(1);}

   // auto-detect num cores by try+fail sched_affinity
   CPU_ZERO(&cpuset); 
   for(j=0,i=CPU_SETSIZE;i;i>>=1){
     CPU_SET(j|i,&cpuset);
     if (!sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) j|=i; // no_err!
     CPU_CLR(j|i,&cpuset);
   }
   if (!mpi_i) printf("# Max-detected-CPU-Cores/Threads= %ld\n", j+1);
   ncores=j+1;
      
   for (ii=0;ii<8*sizeof(long) && ii<mpi_n;ii++) {  // set cmask
     j=((ii/pin.size)*pin.dist
       +(ii%pin.size)+pin.ofs)%pin.ncpu;
     if (j<8*sizeof(long)) cmask |= 1UL<<j;
   }
   if (!mpi_i)
     printf("# CPU-Pinning: idx=((i/%d)*%d+(i%%%d)+%d)%%%d, mask=0x...%08lx\n",
     pin.size, pin.dist, pin.size, pin.ofs, pin.ncpu,
      cmask);
   if (pin.ncpu>ncores && !mpi_i) printf("WARNING: pin.ncpu>ncores\n");

#if 0 // old 32bit/64bit limited to cpus=bits-1
   { int in=0;
     // 2013-05 endless loop for 32 Cores on int32-mask CPUSET=0xffffffff
     // for (ii=0; CFG_CPUSET>>ii && ii<64; ii++) if (1&(CFG_CPUSET>>ii)) ncores++;
     for (ii=0; CFG_CPUSET>>ii && ii<64; ii++) if (1&(CFG_CPUSET>>ii)) {
       if (mpi_i%ncores==in) break; in++; // mycore=0,1,2,..,0,1,2,.. --byslot
     }
   }  // ii = cpucore(mpi_i)
#else  // new method for more than 32 cores
   // does only cover equidistant distribution but most
   // MASK     
   // 0xffff = {16,1,1,0}
   // 0x6666 = {16,2,4,1}
   // 0x0f0f = {16,4,8,0} = ((ii/4)*8+0)%16
   // 0xf0f0 = {16,4,8,4}
   ii=((mpi_i/pin.size)*pin.dist
      +(mpi_i%pin.size)+pin.ofs)%pin.ncpu;
   // ncores=pin.ncpu;
#endif
   if (ii>=CPU_SETSIZE) {printf("ERROR: ii>CPU_SETSIZE, ii=%d\n",ii);exit(1);}
   CPU_ZERO(&cpuset);   // void CPU_ZERO(cpu_set_t *set); bits[idx]|=mask
   CPU_SET(ii,&cpuset); //  ->bits[ii/NCPUBITS] |= (1<<(ii%NCPUBITS))
   
   // The constant CPU_SETSIZE (1024) specifies a value one greater than  the
   //      maximum CPU number that can be stored in a CPU set.
   usleep((mpi_i%16+1)*10000L); // 10ms..170ms, sort outputs
   if (mpi_i<4 || mpi_i>=mpi_n-2) {
     printf("# set cpu affinity for core(%d)=%d size=%d\n",
              mpi_i, ii, CPU_SETSIZE);
   }
   // no sched_setaffinity on OSF1
   if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     printf("# set cpu affinity for core(%d)=%d size=%d failed\n",
             mpi_i, ii, CPU_SETSIZE);
             usleep(100000L);perror("set_affinity");
   }
   usleep(100000); // 100ms, pseudo sync
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  if (!mpi_i) printf("#");
   /* 2013-05 gcc -fopenmp -v ... =
     -L/usr/lib/gcc/x86_64-redhat-linux/4.1.2 -lgomp -lpthread */
#ifdef CFG_OMP
  // omp_set_num_threads(2);
#pragma omp parallel
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  /* next line is outside parallel region */
  if (!mpi_i) printf(" OpenMP threads=%d", omp_n); /* same on every node? */
#endif
#ifdef MPI_VERSION
  if (!mpi_i) printf(" MPI-%d.%d",MPI_VERSION,MPI_SUBVERSION);
#endif
  if (!mpi_i)
    printf(" mpitasks=%d mem/rank= %.1f MB  int%d  mintime/s=%.1f test=%d\n"
           "# measure: BW = bandwidth, aBW = aggregate BW, lat = latency\n",
      mpi_n, (1l<<(log2size-10))/1024., 1<<log2long, tmin, testmode);

  if ((memhole0 || memhole1) && (memhole1<=memhole0 || memhole0<0)){
    printf("ERROR: memhole end<=start or start<0\n");exit(1);}

  if (memhole0 || memhole1){
    long int i1; 
    static char *a1=NULL, *a2=NULL;
    printf("# reserve memory hole (start-end=%ld-%ld)\n", memhole0, memhole1);
    if (memhole1<=memhole0 || memhole0<0){
      printf("ERROR: memhole end<=start or start<0\n");exit(1);}
    a1 = (char *) malloc(          memhole0 <<20);
    a2 = (char *) malloc((memhole1-memhole0)<<20);
    if ((!a1) || !a2) {printf("ERR malloc\n");exit(7);}
    // touch every kB
    for (i1=0;i1<          memhole0 <<10;i1++) a1[i1<<10]=0; // good mem
    for (i1=0;i1<(memhole1-memhole0)<<10;i1++) a2[i1<<10]=0; //  bad mem
    if (a2[0]!=a1[0]) printf("ERROR bad mem is bad, but reserved\n");
    free(a1); // get good mem back, but do not release a2
    printf("# hole reserved, ready for progress ...\n");
    // hopefully compiler does no garbage collection!
  }
                         
  /* old: asize = 1l << log2size */
  /* new: asize = option.asize << log2size 2013-06-14 */
  aa = (long int *) malloc(asize); /* malloc but touch it by threads! */
  { /* MPI_Allreduce !aa and get min/max maxmalloc, also sync 2013-06 */
    int aa1,aa2; long int size1, size2;
    aa1=((aa)?0:1); /* !0 = failure */
    MPI_Allreduce(&aa1, &aa2, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    if (aa2 && !mpi_i) printf("ERROR: malloc failed on %d nodes\n", aa2);
    if (aa1 && aa2<8) printf("INFO: malloc failed on node %d\n", mpi_i);
    if (aa2) { free(aa); aa=NULL; /* get max alloc ToDo! */
      for (size1=0,size2=asize;size2>=1<<20;size2>>=1) { /* log2-test */
        aa = (long int *) malloc(size1+size2);
        if (aa) { size1+=size2; free(aa); aa=NULL; }
      } /* now we have the maxalloc on each node, range = 1MB..2*asize-1 */
      MPI_Allreduce(&size1, &size2, 1, MPI_LONG, MPI_MIN, MPI_COMM_WORLD);
      if (!mpi_i) printf("INFO: min maxmalloc/MB = %ld\n", size2>>20);
      if (aa2<8 && size2==size1) printf("INFO: min on node %d\n", mpi_i);
    }
  }
  /* in MPI failed malloc cause signal 11 ? on CentOS */
  /* in MPI Abort generates lot of error lines Gentoo1.12.11.1 + Slurm1 */
  if (!aa) testmode=0; /* MPI_Abort(MPI_COMM_WORLD, 133); */
  
#if 0
  /* ToDo: compute CPU clock, diff loops over n vs. n+1 ops? */
  if (!mpi_i) {
    unsigned long i0,i1=0,i2,j=7,i6=1000000; double t2,t1,t0,
      tmin2=1.0; /* measuring time in seconds (old: tmin2=tmin) */
    if (argn==7000) i6=100;  /* knocking out the optimizer :) */ 
    /* ToDo: second ident loop for deviation mesurements (t1b,t2b) */ 
    t0=MPI_Wtime();t1=0.0;
    for (i0=1;i0;i0<<=1) { /* double loops until 1s gone, avoid overrun i0 */
      for (;i1<i0;i1++)
        for (i2=0;i2<i6;i2++) { j=j^i2; }
      t1=MPI_Wtime()-t0;
//      printf("# loop1=   %6lde6 / %6.3f s = %9.0fe6/s\n", i0, t1, i0/t1);
      if (t1>tmin2) break;
    }
    printf("# loop1=   %6lde6 / %6.3f s = %9.0fe6/s ^\n", i0, t1, i0/t1);
    t0=MPI_Wtime();t2=0.0;
    for (i1=0;i1<i0;i1++)   // add single ops until time is +10% minimum ??
      for (i2=0;i2<i6;i2++) { j=j^i2^3; } // MIPS-500MHz 256e6/(1.6s-1.1s)
    t2=MPI_Wtime()-t0;
    printf("# loop2=   %6lde6 / %6.3f s = %9.0fe6/s ^^\n", i0, t2, i0/t2);
    if (t2-t1<0.1) { /*  2013-06-18 not for MIPS! */
     t0=MPI_Wtime();t2=0.0;
     for (i1=0;i1<i0;i1++)   // add single ops until time is +10% minimum ??
       for (i2=0;i2<i6;i2++) { j=j^i2^(i2>>7)^3; } // AMD ^,>> parallel!
     t2=MPI_Wtime()-t0;
     printf("# loop2=   %6lde6 / %6.3f s = %9.0fe6/s ^^(>>)\n", i0, t2, i0/t2);
    }
    if (t2!=t1)
    printf("# loop2-1= %6lde6 / %6.3f s = %9.0fe6/s\n", i0,t2-t1,i0/(t2-t1));
    if (j==863479817) fputs("# super hit :)\n",stdout);
    /* 2013-05
     * Intel E5400      2.70GHz loop2-1=  32768e6 /  3.415 s =  9596e6/s ??
     * AMD Opteron 6134 2.30GHz loop2-1=  65536e6 / 28.534 s =  2297e6/s 
     * MIPS 500MHz              loop2-1=    256e6 /  1.613 s =   159e6/s ??
     */
  }
#endif

#if 1
  if (testmode&(16^8^4^2)) {
    /* preinit a_rnd and reuse for speedup, but randomness gets more bad */
    srandom(mpi_i);  // different access patterns on different CPUs
    /* long int random() returns 31bit random only! */
    r00= 0l; /* xor sum */
    for (k=0;k<1024;k++) /* 64bit by 3*31 bit 2013-06-20 */
      r00 ^= a_rnd[k] = random() ^ ((random() ^ (random()<<8))<<31);
    a_rnd[1023] ^= r00;  /* ensure sum_xor a_rnd[k] = 0 */
    r00= 0l; for (k=0;k<1024;k++) r00 ^= a_rnd[k]; /* xor sum */
    if (r00!=0) printf("ERR: rnd sum^!=0 %016lx\n", r00);
    /* replace random() by (a_rnd[i&1023]^a_rnd[(i>>10)&1023]^...>>20...) */
    /* 1024^2 is not enough randomness, for 2^30 size arrays ToDo! */
  }
#endif

  if (testmode&(4^2)) {
    if (!mpi_i)
      puts("# Initialize+check random array for each MemSize (slow) ...");
  }

#if 1
/* ToDo: prefetchable random access, randomness from extra 1k array + xor */
if (testmode&2){ /* prefetchable random access */
  /* 
   * AMD Opteron 6134 2.30GHz 16*4GB aBW.r[i]4p=2200MB/s,59ns r[i]=1000MB/s,127ns
   */
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]\n");
  for (a2=alen;a2>>5;a2>>=1) {
    // its twicky to have a random "next_index" table without shortcuts 
    // we need a 2nd array for prev_index
    for (k=0;k<a2;k++) {       // define a ring, backwards as starting point
      aa[k]=(k+a2-1)%(a2); // successors (old: &(a2-1) only for 2^n)
    }
#pragma omp parallel for default(shared) private(k,i,r00,r01,r10,r11,r02,r12)
    for (k=0;k<a2;k++) {  // shaking the chain, this is slow! a2=arraysize
      // chooze a random pair r0=i, r1=j to exchange i+1 and j+1
      i= a_rnd[k&1023]^a_rnd[(k>>10)&1023]^a_rnd[(k>>20)&1023]; /* random() */
      r00=k;       r10=((unsigned long)i)%a2; // predecessors i, j
      r01=aa[r00]; r11=aa[r10];            // exchange partners i+1, j+1
      r02=aa[r01]; r12=aa[r11];            // successors i+2, j+2
      if (r12==r01 || r10==r01) continue;  // dont exchange neighbours
      // idex:    0   1   2   3   4   5   6
      // succs:   1   2   3   4   5   6   0  0-1-2-3-4-5-6-0
      //        r00 r01 r02 r10 r11 r12
      // exchng:      ^           ^          0-4-2-3-1-5-6-0 (1,4 exchanged)
      // succs:   4   5   3   1   2   6   0
      //          ^   ^       ^   ^   
      aa[r00]=r11; aa[r01]=r12; // new successors of i, i+1
      aa[r10]=r01; aa[r11]=r02; // new successors of j, j+1
      if (alen<=64 && !mpi_i) check_ring(a2,1,mpi_i);
      if (alen<=64 && !mpi_i) printf(" exchange: %2d %2d", (int)r01, (int)r11);
    }
    if (alen<=128*1024 && !mpi_i) check_ring(a2,0,mpi_i);
//    printf("OK %d\n",__LINE__);

    // measure random access = sizeof(long)/memory_latency
    MPI_Barrier(MPI_COMM_WORLD);
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      j=i;
      // ToDo: bad usage if reset ll for every inner loop!?
      do { // repeat until some seconds are gone
         ll=0; k=a2/4;             // gcc-4.1 x86 2 instr.
         r00=0; r01=a2/4; r10=a2/2; r11=a2/4*3;
// ToDo: OMP
         do {
           // 4 independent random access sequences
           r00=aa[r00];  // independent values, 3 of 4 are prefetchable?
           r01=aa[r00];  // independent values, 3 of 4 are prefetchable?
           r10=aa[r10];  // independent values, 3 of 4 are prefetchable?
           r11=aa[r11];  // independent values, 3 of 4 are prefetchable?
           // ll=aa[ll];             // gcc-4.1 x86 1 instr. + latency
           // __builtin_prefetch(aa+l);
         } while(--k);            // gcc-4.1 x86 2 instr.
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f random i=r[i].4p\n",
         ((t2>=tmin)?' ':'#'), mpi_n, 1+0*omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*sizeof(long)*1.*loops/t2,
                          a2*sizeof(long)*1.*loops/t2, t2*1e9/(loops*1.*a2));
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    }
    if (r00==r01 && r10==r11) puts("??\n"); /* dummy usage */
  } /* for (a2=alen;a2>>5;a2>>=1) down to 2**5=32B */
}


if (testmode&4){ /* non-prefetchable random access */
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]\n");
  for (a2=alen;a2>>5;a2>>=1) {
    // its twicky to have a random "next_index" table without shortcuts 
    // we need a 2nd array for prev_index
    for (k=0;k<a2;k++) {       // define a ring, backwards as starting point
      aa[k]=(k+a2-1)%a2; // successors
    }
    for (k=0;k<a2;k++) {  // shaking the chain, this is slow!
      // chooze a random pair r0=i, r1=j to exchange i+1 and j+1
      i= a_rnd[k&1023]^a_rnd[(k>>10)&1023]^a_rnd[(k>>20)&1023]; /* random() */
      /* long % 2M may result to 0xffffffff8768 for negative longs */
      r00=k;       r10=((unsigned long)i)%a2;  // predecessors i, j
      r01=aa[r00]; r11=aa[r10];            // exchange partners i+1, j+1
      r02=aa[r01]; r12=aa[r11];            // successors i+2, j+2
      if (r12==r01 || r10==r01) continue;  // dont exchange neighbours
      // idex:    0   1   2   3   4   5   6
      // succs:   1   2   3   4   5   6   0  0-1-2-3-4-5-6-0
      //        r00 r01 r02 r10 r11 r12
      // exchng:      ^           ^          0-4-2-3-1-5-6-0 (1,4 exchanged)
      // succs:   4   5   3   1   2   6   0
      //          ^   ^       ^   ^   
      aa[r00]=r11; aa[r01]=r12; // new successors of i, i+1
      aa[r10]=r01; aa[r11]=r02; // new successors of j, j+1
      if (alen<=64 && !mpi_i) check_ring(a2,1,mpi_i);
      if (alen<=64 && !mpi_i) printf(" exchange: %2d %2d", (int)r01, (int)r11);
    }
    if (alen<=128*1024 && !mpi_i) check_ring(a2,0,mpi_i);

    // measure random access = sizeof(long)/memory_latency
    MPI_Barrier(MPI_COMM_WORLD);
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      j=i;
      do { // repeat until some seconds are gone
         ll=0; k=a2;             // gcc-4.1 x86 2 instr.
         do {
           ll=aa[ll];             // gcc-4.1 x86 1 instr. + latency
           // __builtin_prefetch(aa+l);
         } while(--k);            // gcc-4.1 x86 2 instr.
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f random i=r[i]\n",
         ((t2>=tmin)?' ':'#'), mpi_n, 1+0*omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*sizeof(long)*1.*loops/t2,
                          a2*sizeof(long)*1.*loops/t2, t2*1e9/(loops*1.*a2));
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    }
  } /* for (a2=alen;a2>>5;a2>>=1) down to 2**5=32B */
}  /* non-prefetchable random access */

  r00=ll; // avoid optimizations which skip the loop

  /* 
   * AMD Opteron 6134 2.30GHz 16*4GB lstream.aBW=46.7GB/s,2.7ns p4=50GB/s,2.54ns
   */
if (testmode&1){ /* sequential read = lstream */
  MPI_Barrier(MPI_COMM_WORLD);
  if (!mpi_i)
    printf("# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]\n");
  // measure memory throughput for step0
  for (a2=alen;a2>>7;a2>>=1) { // reduce msgsize to 1<<7=128 * 8 = 1k..2kB
    long int *ap=aa;  // moving pointer    
    long int x1, x2, x3, x4; /* decoupled 4 subresults */
    //    double t3;
    loops=0; x1=x2=x3=x4=0;
#pragma omp parallel for default(shared) private(k)
    for (k=0;k<a2;k++) {       // initialize mem (using thread affinity)
      aa[k]=-(k&1);            // 00 ff pattern
    }
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) { // repeat until mintime
      j=i;
      do { // repeat until some seconds are needed
         long int *ae=aa+a2;  // 2013-10-30
         // echo 1200000 >/sys/dev*/system/cpu/cpu0/cpufreq/scaling_max_freq
         //  k=a2/4;do{ ... }while(--k); # t=1.92ns(256M) 0.84(10k) 0.88(1k)
         //  for(k=0;k<a2;k+=4){ ... };  # t=1.95ns(256M) 0.84(10k) 0.88(1k)
         // omp1(k=0;k<a2;k+=4){ ... };  # t=1.96ns(256M) 1.20(16k) 6.64(1k)
         //  manual unroll has no effect on 2700MHz, but 1200MHz???
         //  for(k=0;k<a2;k+=4) # t=1.54ns(256M) PC2700MHz
         //  for(k=0;k<a2;k+=1) # t=1.51ns(256M) 0.58(32k) 0.70(1k) PC2700MHz
         //  for(k=0;k<a2;k+=1) # t=1.97ns(256M) 1.24(32k) 1.57(1k) PC1200MHz
         // ToDo: ... -Os
#pragma omp parallel for default(shared) private(k,ap) reduction(^:x1,x2,x3,x4) firstprivate(aa,a2)
// ToDo: all threads start with ap = bad! missing const directive a2
// lot of cpu overhead (16-32x) for openmp + small arrays (small a2)! cache coh.
//   -Os besser als -O2 !?
#if 0    // slower on cache (fast ram)
         for (k=0;k<a2;k+=1) {  /* down=2.08-2.13ns, up=2.16-2.20ns */
           x1^=aa[k]; // 1.51ns(256M,PC2700MHz) 1.97ns(256M,PC1200MHz)
                      // 1.97ns(8M) 1.53ns(1M) 1.22ns(16k) gcc44.Os=O3 PC1200
                      //   erreicht bei L1-cache 0.58ns = 1.566clks = ...
#else   // mem-bustakt? DDR800(400MHz)*4B? memtest=2996MB/s 4B/1.54ns=2.6GB/s
         for (k=0;k<a2;k+=4) {  /* down=2.08-2.13ns, up=2.16-2.20ns */
           x1^=aa[k  ]; //     0.833ns = 1/1200MHz (5/4*0.833ns=1.04ns)
           x2^=aa[k+1]; // 1.96ns(8M) 1.53ns(1M) 0.84ns(16k) gcc44.O3=Os PC1200
           x3^=aa[k+2]; // 1.96ns(gcc44,256M,PC1200MHz -Os=O2=O3=gcc41.O3)
           x4^=aa[k+3]; // 1.54ns(256M,PC2700MHz) 2.07ns(256M,PC1200MHz -Os)
           // gcc44O3openmp 1task: 1.96ns(256M) 4.1GB/s 1.20ns(16k)  6ns(1k) 1200MHz
           // gcc44O3openmp 2task: 1.79ns(256M) 4.5GB/s 1.18ns(16k) 12ns(1k) 1200MHz
           //  Beschleunigung kaum sichtbar! warum nicht? mem=3GB/s L1=44GB/s
           //    mem long/BW=8B/3000MB/s=2.6ns  nax=800MHz*8B=6.4GB/s
           // openmp overhead ab 16k sichtbar, unter 4k dominant
#endif
//   nicht besser mit pointer ap=aa+k, da openmp int braucht, hier nicht genutzt
// __builtin_prefetch(aa); // gcc-4.1 x86-msse2 2 instr.
// schlechter fuer kleine arrays wegen coherence tests!
         }
         /* ToDo: j loop is bad for x1^x2... result! */
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      xy=x1^x2^x3^x4;
      if (xy!=0) { /* vor MPI_Reduce! */
        printf("ERR: bad result x=%016lx %s\n", xy,
         getenv("SLURMD_NODENAME")); // no HOSTNAME
        // system("hostname"); // 2013-06-18 kautz, failes wenn OOM
      }
      /* Bcast laeuft auch durch L2-ECC-gecrashte kautz-nodes! 2013-06 */
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD); /* old */
      /* MPI_Reduce bringt komische Resultate fuer loops, t2 passt nicht? */
      // t3=t2;MPI_Reduce(&t3, &t2, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>10) && !mpi_i) // alive and final message
      printf("%c%3d %2d %8.2e %9ld %5.2f %8.2e %8.2e %6.2f lstream\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n, a2*1.*sizeof(long int),
         loops, t2, mpi_n*a2*1.*sizeof(long int)*loops/t2,
                          a2*1.*sizeof(long int)*loops/t2,
                t2*1e9/(loops*1.*a2));
      // l must be used, some compilers remove the mem access else
      if (t2>=tmin) break;
    } /* repeat until mintime */
  }
} /* sequential read */

/* ToDo: STRESS = endless write/read (count writes, reads) report bitflips */
/* numademo+memtest86+v4.20 is bad stresstest because single thread
 *   using mpirun -np 16 numademo? 
 *  8: 00    00    00    00   vs. ff    ff    ff    ff 
 * 16: rnd rnd^~0 rnd rnd^~0  vs. rnd^~0 rnd rnd^~0 rnd
 */
if (testmode&(8^16)){
  /* new alloc to allow locality for OMPI (could be done at begin!) */
  /* ToDo OMP, old: ompi only, use double + FPU for PSU test
   * comp2+kW? + check result (store first to source and compare later)
   * comp2: 2*(8*2300MHz,64GB)  numademo4n8GB: rnd=90MB/s memcpy=2-6GB/s 200m++
   *   numademo wandert zw. cores, use taskset 0x01 ... ? wird ignoriert!
   *     numactl --physcpubind 0 wird ab running on node 1 ignoriert!
   *   numademo  8m stream random #  24s str=3-6GB/s rnd=200-300MB/s
   *   numademo 16m stream random #  57s str=3-5GB/s rnd=170-260MB/s
   *   numademo 80m stream random #      str=3-5GB/s rnd=150-220MB/s
   *  mpicc -O2 -o memspeed -DCFG_MPI -DCFG_CPUSET=0xffff ../utils/memspeed.c 
   *  mpirun -np 16 ./memspeed --log2size=30 --mintime=2 --test=8 # 
   * 2013.05: 16*1GB 16*2925MB/s (no cpuset), 16*3760MB/s (cpuset16) 60GB/s
   *          16*2GB 16*3751GB/s (37GBcached == drop_caches)
   *          16*4GB 16*3740GB/s # node04: 3737MB/s, node19: 
   */
  int tmode=16, cnt=0;
  if (!mpi_i)
    printf("# R/W+Power-Stress + check for bitflips (mode=%d)\n"
   "# mpi omp Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]\n", testmode);
#pragma omp parallel for default(shared) private(k) firstprivate(aa,alen)
  for (k=0;k<alen;k++) aa[k]=-(k&1);   // initialize mem 00 ff pattern
  // measure memory throughput for step0
  for (cnt=0,a2=alen;;cnt++) { // do not reduce msgsize, stress it endless! alen=1<<alog
    long int x1, x2, *ap=aa, ae=0l, ae2;  // moving pointer
    /* rot through bit 8 + 16 until 1 bit found */
    for(;;) {
      tmode=((tmode<<1)&0x18) | ((tmode&0x10)>>1);
      if (tmode&testmode) break;
    }

    ae=a_rnd[cnt&1023]^a_rnd[(cnt>>10)&1023];  /* rnd(1024^2), start=0 */
    if (tmode&8) ae=0l; /* 00 ff sequence */
    
    loops=0; x1=x2=xy=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
    register double r1=0.12, r2=0.12, r3=0.13, r4=0.14; // FLOPS to stress PDU
    register double r5=0.22, r6=0.22, r7=0.23, r8=0.24;
    double r9=0.31;
      j=i; // 2013 comp2: first round runs slower, autotune cache method?
      // 1st_round=393MB/s, 2nd++_round=1610MB/s (faktor4), training? throttle?
      do { // repeat until some seconds are reached
         ap=aa; k=a2/2; ae^=~0l;
         ae2= ((tmode&16)?ae^~0l:ae); // gcc-4.1 x86 2 instr.
         // k=a2/2;do{...}while(--k);  # 8.49ns(256MB) 0.94GFLOPs(19.6%)
         // for(k=0;k<a2;k+=2){...}    # 8.69ns(256MB) 0.92GFLOPs(19.2%)
         // omp1(k=0;k<a2/2;k++){...}  # 4.52ns(256MB) 1.77GFLOPs(36.9%) 1.2GHz
         // 1200MHz*4FLOP=4.8GFLOPs (8FLOP/1.67ns) 
        // Pentium(R) Dual-Core CPU E5400 @ 1200-2700MHz
// ToDo: bad pragma-option for r1-r8 (reduction is bad)
// ToDo: ap[] access is all threads to same memory == bad!
#pragma omp parallel for default(shared) private(k) reduction(^:x1,x2) firstprivate(aa,a2,ae,ae2) reduction(+:r1,r2,r3,r4,r5,r6,r7,r8)
         for(k=0;k<a2;k+=2) {  // 2013-09-11 // 3 instr.
//         do {
//           ap=aa+k;
// omp.L330 8*mulsd %xmmX,%xmmX + 8*addsd %xmmY, %xmmX (16 instr.) 2FLOPs/clk
// gcc.L207 8*(2*movsd+mulsd+addsd) from/to memory     (32 instr.) 1FLOP/clk
// ToDo: vergleich auf moderner CPU!
           r1=r1*r1+0.1;    // cpu burn code (stress the system) 8 FLOPs/cycle
           r2=r2*r2+0.2;
           r3=r3*r3+0.3;
           r4=r4*r4+0.4;
           r5=r5*r5+0.1;
           r6=r6*r6+0.2;
           r7=r7*r7+0.3;
           r8=r8*r8+0.4; // without next 4 lines 8FLOP/0.21ns=38GFLOPs @ 1.2GHz?
//                 gcc-O1 8FLOP/5.44ns=1.47GFLOPs(31%*4.8GF) (faster!?)
//                 gcc-O2 8FLOP/7.10ns=1.13GFLOPs(23%*peak4.8GF)
//                 omp1-O2 8FLOP/6.91ns=1.16GFLOPs (shared rx)
//                 omp1-O2 8FLOP/3.34ns=2.40GFLOPs(50%*4.8GF) (reduction rx)
//          t2=MPI_Wtime()+MPI_Wtime(); // localisation in .s
#if 0
           x1^=*(ap  );           // gcc-4.1 x86 1 instr. + latency
           x2^=*(ap+1);
           *(ap  )=ae;
           *(ap+1)=ae2;           //   ... new pattern write
           ap+=2;                 // gcc-4.1 x86 1 instr., bad for openmp!
           // gcc44-O3 4.57ns PC1200MHz
#else 
           x1^=aa[k];           // gcc-4.1 x86 1 instr. + latency
           x2^=aa[k+1];
           aa[k  ]=ae;
           aa[k+1]=ae2;           //   ... new pattern write
           // gcc44-O3 4.52ns PC1200MHz (4.58ns 2 threads! slower?!)
           //  read + write max reached?  3.5GB/s(r+w)
           // quanta2: 9GB/s(1) 18.2GB/s(2) 36GB/s(4) (68GB/s)(2.6GB/8)
           //     104GB/s(4GB/16)  97GB/s(6GB/32) 96GB(80GB/64)
           // OMP_NUM_THREADS=16 nice -19 ./memspeed --test=8 --asize=750000
           //   750GB 7.11. 18:18 ECC-Fehler! ^c
           //   370GB 18:19 ECC-Fehler
           //   200GB 18:20 ECC-Fehler
           //   100GB 18:21 ECC-Fehler 104GB/s, 50GB 18:21:54
           //  10GB 82.9GB/s ohne Fehler 16MB cache
           //  20GB 104GB/s mit Fehlern 18:25:14
           //  15GB  81GB/s mit Fehlern 18:25:54
           //  12GB 105GB/s mit Fehlern 18:26:55
           //    11GB 105GB/s mit Fehlern 18:27:42
           //    10GB 105GB/s mit Fehlern 18:28:25
           
           //     8GB 103GB/s mit Fehlern 18:29:25
           //     # ToDo einzelne CPUs testen (numactl) ! works test=8
           //   numactl --hardware # --physcpubind 0-7 --membind=0 26.7GB/s
           //     1cpu=9GB/s(remote=5.0GB/s,remote2=3.8GB/s) 4nodes=105GB/s
           //  ToDo: --memhole=10G-11G (alloc in 1GB + touch, free)
           //      --test=idle
#endif
           // __builtin_prefetch(aa); // gcc-4.1 x86-msse2 2 instr.
           // ToDo: better prefetch every 4th access only + unroll loop
//         } while(--k);            // gcc-4.1 x86 2 instr.
         } // for(k=0..a2/2)
         // ToDo: problem x1 x2 ^= over multiple loops hides bitflips!
      } while(--j);               // gcc-4.1 x86 2 instr.
      loops+=i;
      t2=MPI_Wtime()-t1;
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>60*60) && !mpi_i) // alive and final message
      /* aBW= 1.02e+06 %8.2e 1%, 2013-06-20 fix bad factor 2 Mem/Task */
      printf("%c%3d %2d %8.2e %9ld %6.2f %8.2e %8.2e %6.2f lr+w4*+%02d\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n, a2*1.*sizeof(long int),
         /* factor 2 for r+w */
         loops, t2, mpi_n*a2*2.*sizeof(long int)*loops/t2, /* r+w = 2x */
                          a2*2.*sizeof(long int)*loops/t2,
                t2*1e9/(loops*1.*a2), tmode);
      // comp2: 2*1.6GB/s = 2*8B*200MHz=2*8B*2.3GHz/11.5clks (7used?)
      // l must be used, some compilers remove the mem access else
      xy=x1^x2;
      if (xy!=0lu)
//       printf("WARNING: xy= %016lx bitflip?\n", xy);
        printf("ERR: bad result rnd=%016lx x=%016lx %s\n", ae, xy,
         getenv("SLURMD_NODENAME")); // no HOSTNAME
      if (r1==0.6 || r2==0.4 || r3==0.5 || r4==0.23) printf("??\n");
      if (r9==3.4 || r8==0.6 || r5==0.4 || r6==0.5 || r7==0.2) printf("??\n");
      if (t2>=tmin) break;
    }
  }
}

#endif

#ifndef NA
#define NA (8)  /* vector array */
#endif
/* ToDo: FLOAT64-STRESS = max power consumption 2013-09-12 */
/* if (testmode&32) */{
      double ra[NA] __attribute__ ((aligned (16))),
             rb[NA] __attribute__ ((aligned (16))),
             rc[NA] __attribute__ ((aligned (16)));
      double r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11;
  int tmode=32, cnt=0;
  if (!mpi_i)
    printf("# FLOAT64-Power-Stress (mode=%d, 2FLOP/clk @ X5650)\n"
   "# mpi omp loops t[s]  aFLOPS[GF/s] FLOPS[GF/s] lat[ns]\n", testmode);
  // measure memory throughput for step0
  for (cnt=0,a2=alen;;cnt++) { // do not reduce msgsize, stress it endless! alen=1<<alog
    /* rot through bit 8 + 16 until 1 bit found */
    a2=64000000; /* fix for unroll */
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      double ry;
      int k2, k3;
      for (k2=0;k2<NA;k2++) rc[k2]=rb[k2]=ra[k2]=0.001*k2+0.007; // nonzero! 
      ry=0.43;
      r0=0.2;r1=0.3;r2=0.4;r3=0.5;r4=0.6;r5=0.7;
      r6=0.8;r7=0.9;r8=1.0;r9=1.1;r10=1.2;r11=0.1;
      j=i; // 2013 comp2: first round runs slower, autotune cache method?
      // 1st_round=393MB/s, 2nd++_round=1610MB/s (faktor4), training? throttle?
      do { // repeat until some seconds are reached
         // k=a2/2;do{...}while(--k);  # 8.49ns(256MB) 0.94GFLOPs(19.6%)
         // for(k=0;k<a2;k+=2){...}    # 8.69ns(256MB) 0.92GFLOPs(19.2%)
         // omp1(k=0;k<a2/2;k++){...}  # 4.52ns(256MB) 1.77GFLOPs(36.9%) 1.2GHz
         // 1200MHz*4FLOP=4.8GFLOPs (8FLOP/1.67ns) 
        // Pentium(R) Dual-Core CPU E5400 @ 1200-2700MHz
// ToDo: bad pragma-option for r1-r8 (reduction is bad)
// reduction not possible on arrays!
// omp.a3 must be private for correct results!
#if 0
#pragma omp parallel for default(shared) private(k,k2,k3) firstprivate(a2,ry,rb,rc) reduction(+:r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11)
         for(k3=0;k3<a2;k3+=(6000*NA)) {  // needs: -O3 -funroll-loops
         for(k=0;k<(6000*NA);k+=(2*NA)) {  // 2013-09-11 // 3 instr.
#else
#pragma omp parallel for default(shared) private(k,k2) firstprivate(a2,ry,rb,rc) reduction(+:r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11)
         for(k=0;k<a2;k+=(2*NA)) {{  // 2013-09-11 // 3 instr.
#endif
// omp-4.1.2.L330 12*mulsd 12*addsd          0.8FLOP/clk
//       -O1  12*(movsd+mulsd+addsd)         0.9FLOP/clk
// gcc-4.1.2.L211 12*mulsd 12*addsd          1.0FLOP/clk
//       -O3.L241 12*mul 12*add 2*mov        1.7FLOP/clk
//       -O3.L241 8*(mov+mul+add+mov)        2.0FLOP/clk
//       -O2.L241 8*(mov+mul+mov+add+mov)    1.0FLOP/clk
//       -O2.L241 6(=+=*=)6(=*=+=)=12*+      1.7FLOP/clk
//       -O3 (+6*5+2*+*+=*+*+*++)=12+-       2.0FLOP/clk
//   mulsd + addsd for max. 2FLOP/clk -msse
//   mulpd + addpd for max. 4FLOP/clk (packed double = 2*64bit) _mm_mul_pd()
//   _mm256_mul_pd for max. 8FLOP/clk + gcc4.6 -mavx(intel) vaddpd
// ToDo: see stackoverflow sse simd ymm-registers4*8B r[i]=(a[i]+b[i])*c[i];
//     double a[N] __attribute__ ((aligned (16)));
//     gcc -O3 -msse2 [-mavx]  # (vector: 2x [4x]) * (add+mul: 2x)
//   movapd addpd mulpd movapd OR vmovapd vaddpd vmulpd vmovapd (*pd failes)
//    openmp outer loop, not fixed inner loop
// pse.gcc446-O2 5.25GFOPs 2.67GHz 2.0FLOP/clk (1.6-2.67GHz)
// gcc-4.4.7-O3 
// cpu burn code (stress the system) theory = 4 FLOP/clk
// mulsd has 5 clks latency! so we need minimum 5 independent operations
// ToDo: vergleich auf moderner CPU! better 12 (multiple of 2,3,4,6)
//   E5-4620 + http://linux.web.cern.ch/linux/devtoolset/ gcc-4.7.2
//     scl enable devtoolset-1.1 bash
//     gcc -O3 -DNA=32 -funroll-loops -mavx # 8 FLOP/clk*2.2GHz E5-4620 
#if 1
          for (k2=0;k2<NA;k2++)  // -funroll-loops * vectorize? = (x2 or x4)
            ra[k2]=(ra[k2]+rb[k2])*rc[k2]; // gcc44 -msse2 -mavx
#else          
           r0 *= r0;   r1 += ry;   // block 1
           r2 *= r2;   r3 += ry;   
           r4 *= r4;   r5 += ry;   
           r6 *= r6;   r7 += ry;   
           r8 *= r8;   r9 += ry;   
           r10*= r10;  r11+= ry;   
           r1 *= r1;   r0 += ry;   // block 2
           r3 *= r3;   r2 += ry;
           r5 *= r5;   r4 += ry;
           r7 *= r7;   r6 += ry;
           r9 *= r9;   r8 += ry;
           r11*= r11;  r10+= ry;
#endif
         }} // for(k=0..a2, k+=2)
         // ToDo: problem x1 x2 ^= over multiple loops hides bitflips!
      } while(--j);               // gcc-4.1 x86 2 instr.
#if 0
         rb[0]=r0+r1+r2+r3+r4+r5+r6+r7+r8+r9+r10+r11;
#else
         for (k2=0;k2<NA;k2++) rb[k2]=ra[k2];
#endif
      loops+=i;
      t2=MPI_Wtime()-t1;
      if (((int)t2)>123456789) break;  // easier to find 1234567 at .s code
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>60*60) && !mpi_i) // alive and final message
      /* aBW= 1.02e+06 %8.2e 1%, 2013-06-20 fix bad factor 2 Mem/Task */
      printf("%c%3d %2d %9ld %6.2f %8.2e %8.2e %6.2f %d*+%02d\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n,
      /* factor 4*+ ToDo: aSPEED=threads*loops/maxtime SPEED=loops/mintime */
         loops, t2, mpi_n*a2*1.*loops/t2, a2*1.*loops/(t2*omp_n),
                omp_n*t2*1e9/(loops*1.*a2), NA, tmode);
      for (k2=0;k2<NA;k2++) if (rb[k2]==0.6) puts("??");
      if (t2>=tmin) break;
    }
  }
}

  MPI_Finalize();
  return xy+r00;
}
